"""
TODOs: 
    - experiment with various biases for each individual gate. Example: gate for
      bypass opened 50%, etc
    
    - dropout on inputs might force the GateLayer to learn relevant features. Apply
      different dropouts for different gates
    
    - experiment with different gating FC and different gating inputs (maybe 
      gate dependent more than just inputs)
    
    - replace last gate with residual connection ?
    
    - add layer self-analysis (explain each gate learned features)


Original text:

Multi-Gated Layer

pass_through = linear_nobias(input)
dense_bn_post = bn(f(linear(input)) OR layer_norm(....
dense_bn_pre = f(bn(linear(input))

BN = bn_pos * dense_bn_pre + (1 - bn_pos) * dense_bn_post

out1= has_bn * BN + (1- has_bn) * dense

final = uses_layer * out1 + (1- uses_layer) * pass_through


"""

import numpy as np
import tensorflow as tf


__VER__ = '0.3.6.1'


def get_input_shape(lyr):
  ## assume batch first
  return tuple(lyr.input.shape)[1:]

def explain_model(model, log, np_input):
  _layers = model.layers
  np_temp_input = np_input
  log.P("Self-explainability of MGU layers in {} layer graph".format(
    len(_layers)), 
    color='y')
  for _layer in _layers:
    if type(_layer) is MultiGatedUnit:
      _input_shape = tuple(_layer.input.shape)[1:]
      if np_temp_input.shape[1:] == _input_shape:
        # 1. create model from tf_input up to current MGU inpu
        # 2. run inference on np_input
        # 3. pass np_input to current MGU explain
        _layer.explain(
          log=log,
          np_input=np_temp_input,
          )
        _temp_model = _layer.get_model()
        np_temp_input = _temp_model.predict(np_temp_input)
      else:
        log.P("  Layer {} skipped - input {} vs layer input {}".format(
          _layer.name,
          np_input.shape[1:],
          _input_shape,
          ), color='y')


class GatingUnit(tf.keras.layers.Layer):
  def __init__(self, 
               layer,
               activation='sigmoid',
               **kwargs):
    self.gate_trans = layer
    
    self._name = kwargs.get('name', 'gu')
    # replaced tf.keras.activations.get(activation)
    self.gate_activ = tf.keras.layers.Activation(
      activation, 
      name=self._name + '_gact',
      )
    
    super().__init__(**kwargs)
    return
    
    
  def call(self, inputs):
    """
    

    Parameters
    ----------
    inputs : tuple(3) tensors
      inputs[0] - initial input in the module that generates the gate status
      inputs[1] - first path (gate open)
      inputs[2] - second path (gate closed)

    Returns
    -------
    tf_out : TYPE
      DESCRIPTION.

    """
    return self._forward(inputs)
  ############################################################################

  def _forward(self, inputs, for_model=False):
    tf_source = inputs[0]
    tf_value1 = inputs[1]
    tf_value2 = inputs[2]
    
    tf_gate_x = self.gate_activ(
      self.gate_trans(
        tf_source
        ))
    
    if not for_model:
      tf_out = tf_gate_x * tf_value1 + (1 - tf_gate_x) * tf_value2
    else:
      tf_v1 = tf.keras.layers.multiply(
        [tf_gate_x,tf_value1], 
        name=self._name + '_open_go_s1')
      tf_not_gate_x = tf.keras.layers.Lambda(
        lambda x: 1-x, 
        name=self._name + '_neg_gate',
        )(tf_gate_x)
      tf_v2 = tf.keras.layers.multiply(
        [tf_not_gate_x,tf_value2], 
        name=self._name + '_close_go_s2')
      tf_out = tf.keras.layers.add(
        [tf_v1, tf_v2],
        name=self._name + '_add')
    return tf_out
  
  
  def _get_gate_response_tf(self, tf_source):
    return self.gate_activ(
      self.gate_trans(
        tf_source
        ))
  
  def _get_gate_response_model(self):
    input_shape = get_input_shape(self.gate_trans)
    tf_inp = tf.keras.layers.Input(input_shape)
    tf_x = self._get_gate_response_tf(tf_inp)
    m = tf.keras.models.Model(
      inputs=tf_inp,
      outputs=tf_x,
      )
    return m


class SqueezeGatingUnit(tf.keras.layers.Layer):
  def __init__(self,
               input_shape,
               output_units,
               base_layer_activation,
               activation,
               name,
               squeeze_type,
               **kwargs):

    assert base_layer_activation not in [None, 'linear']
    nr_dim = len(input_shape[1:])
    input_units = input_shape[-1]
    self.squeeze_type = squeeze_type
    self.prepare = None
    self.out_shape = None
    self.embed = None
    self.expand = None
    self.activate = None
    self.reshape = None

    if nr_dim == 4:
      self.prepare = tf.keras.layers.GlobalAveragePooling3D(name=name + '_gap3d')
      self.out_shape = (1,1,1,output_units)
    elif nr_dim == 3:
      self.prepare = tf.keras.layers.GlobalAveragePooling2D(name=name + '_gap2d')
      self.out_shape = (1,1,output_units)
    elif nr_dim == 2:
      self.prepare = tf.keras.layers.GlobalAveragePooling1D(name=name + '_gap1d')
      self.out_shape = (1,output_units)
    elif nr_dim == 1:
      self.prepare = tf.keras.layers.Lambda(lambda x: x, name=name + '_ident')
      self.out_shape = (output_units,)
    
    self.embed = tf.keras.layers.Dense(
      units=self._infer_embed_units(
        input_units=input_units, 
        adaptive=squeeze_type == 'adaptive'
        ),
      activation=base_layer_activation,
      name=name + '_embed_lin_'+str(base_layer_activation)
    )

    self.expand = tf.keras.layers.Dense(units=output_units, name=name+'_expand')
    self.activate = tf.keras.layers.Activation(activation=activation, name=name+'_activ')
    self.reshape = tf.keras.layers.Reshape(self.out_shape, name=name+'_reshape')
    super().__init__(**kwargs)
    return

  def call(self, inputs):
    return self._forward(inputs)

  @staticmethod
  def _infer_embed_units(input_units, adaptive=True):
    if adaptive:
      stages  = [ 32, 128, 512]
      factors = [  4,   8,  16] 
      assert len(stages) == len(factors)
      factor = 2
      for i in range(len(stages)):
        if input_units > stages[i]:
          factor = factors[i]    
      embed_units = max(input_units // factor, 2)
    else:
      embed_units = max(input_units // 16, 2)
    return embed_units

  def _forward(self, inputs, for_model=False):
    tf_source = inputs[0]
    tf_value1 = inputs[1]
    tf_value2 = inputs[2]

    tf_gate_x = self.prepare(tf_source)
    tf_gate_x = self.embed(tf_gate_x)
    tf_gate_x = self.expand(tf_gate_x)
    tf_gate_x = self.activate(tf_gate_x)
    tf_gate_x = self.reshape(tf_gate_x)

    if not for_model:
      tf_out = tf_gate_x * tf_value1 + (1 - tf_gate_x) * tf_value2
    else:
      tf_v1 = tf.keras.layers.multiply(
        [tf_gate_x,tf_value1],
        name=self._name + '_open_go_s1')
      tf_not_gate_x = tf.keras.layers.Lambda(
        lambda x: 1-x,
        name=self._name + '_neg_gate',
        )(tf_gate_x)
      tf_v2 = tf.keras.layers.multiply(
        [tf_not_gate_x,tf_value2],
        name=self._name + '_close_go_s2')
      tf_out = tf.keras.layers.add(
        [tf_v1, tf_v2],
        name=self._name + '_add')
    return tf_out

  def _get_gate_response_tf(self, tf_source):
    tf_gate_x = self.prepare(tf_source)
    tf_gate_x = self.embed(tf_gate_x)
    tf_gate_x = self.expand(tf_gate_x)
    tf_gate_x = self.activate(tf_gate_x)
    tf_gate_x = self.reshape(tf_gate_x)
    return tf_gate_x

  def _get_gate_response_model(self):
    input_shape = get_input_shape(self.prepare)
    tf_inp = tf.keras.layers.Input(input_shape)
    tf_x = self._get_gate_response_tf(tf_inp)
    m = tf.keras.models.Model(
      inputs=tf_inp,
      outputs=tf_x,
    )
    return m

class MultiGatedUnit(tf.keras.layers.Layer):
  """
  MultiGatedUnit implements a universal advanced version of highway networks that can be applied
  to almost any layer as a wrapper.

  The usage is as simple as:
    ```
    # example 1
    # here we apply a activation without specifying one in base layer - the Conv1D is linear
    lyr_m1 = MultiGatedUnit(tf.keras.layers.Conv1D(64,3), activation='relu')

    # example 2
    # we specify the activation in the FC layer
    lyr2 = tf.keras.layers.Dense(10, activation='sigmoid')
    # we wrap it inside the MultiGatedUnit
    lyr_m2 = MultiGatedUnit(lyr2)
    ```
  """
  def __init__(self,
               layer,
               gates_initializer='glorot_uniform',
               bypass_initializer='glorot_uniform',
               gating_activation='sigmoid',
               add_residual_connection=True,
               use_squeeze_gating='adaptive', # None/False, 'adaptive', 'fixed'
               version=None,
               **kwargs,
               ):
    assert isinstance(layer, tf.keras.layers.Layer), "`layer` must be a `tf.keras.layers.Layer` not `{}`".format(
      layer.__class__.__name__)
    assert use_squeeze_gating in [None, False, 'adaptive','fixed'], "`use_squeeze_gating` must be either disabled or fixed or adaptive based on input size"
    if version is None:
      version = __VER__

    self.__version__ = version
    self.version = self.__version__
      
    self.gates_initializer = gates_initializer
    self.bypass_initializer = bypass_initializer
    self.gating_activation = gating_activation
    self.layer_class = layer.__class__
    self.add_residual_connection = add_residual_connection
    self.use_squeeze_gating = use_squeeze_gating
    
    base_layer_config = layer.get_config()
    name = kwargs.get('name')
    base_name = base_layer_config.get('name','')
    if name is None:
      name = 'MGU_' + base_name

    kwargs['name'] = name    
    if base_name[:3] != 'MGU':
      base_layer_config['name'] = 'MGU_' + base_name + '_base'
    
    self._name = name
    self._original_layer = layer
      
    self.activation = kwargs.get('activation') if 'activation' in kwargs else base_layer_config.get('activation')
    if base_layer_config.get('activation') not in [None, 'linear']:
      base_layer_config['activation'] = 'linear'
      
    if self.activation == 'linear':
      raise ValueError('Cannot have MGU with linear activation. Either set host MGU or the layer activation')

    if 'activation' in kwargs:
      kwargs.pop('activation')
    
    self.base_layer_config = base_layer_config

    ########## CLASS ATTRIBUTRES #############
    self.layer = None
    self.act_pre_bn = None
    self.act_post_bn = None
    self.bypass = None
    self.bn_pre_act = None
    self.bn_pos_act = None
    self.ln_pos_act = None
    self.residual = None

    self.g_bpre_bpos = None
    self.g_bn_ln = None
    self.g_norm_non = None
    self.g_residual = None
    self.g_proc_skip = None
    ##########################################

    self._create_layers()

    super().__init__(**kwargs)
    return

  def call(self, inputs):
    return self._forward(inputs)

  def build(self, input_shape):
    super().build(input_shape)
    self._create_all_gates(input_shape)
    self.built = True
    return

  def get_config(self):
    config = {
        'layer' :  tf.keras.layers.serialize(self.layer),
        # no tf.keras.activations.serialize for string props
        'activation' : self.activation,
        'gating_activation': self.gating_activation, 
        'gates_initializer': self.gates_initializer, 
        'bypass_initializer': self.bypass_initializer,
        'version' : self.version,
        'add_residual_connection' : self.add_residual_connection,
        'use_squeeze_gating' : self.use_squeeze_gating
    }
    base_config = super().get_config()
    cfg =  dict(list(base_config.items()) + list(config.items()))
    return cfg

  @classmethod
  def from_config(cls, config, custom_objects=None):
    from tensorflow.python.keras.layers import deserialize as deserialize_layer  # pylint: disable=g-import-not-at-top
    # Avoid mutating the input dict
    config = config.copy()
    v = config.get('version', 'UNK')
    print("\x1b[1;33mLoading MGU from {} config into v{}\x1b[0m".format(
      v, __VER__))
    layer = deserialize_layer(
        config.pop('layer'), custom_objects=custom_objects)
    return cls(layer, **config)  
  
  #############################################################################


  def _create_layer(self, name, kernel_initializer, use_bias):
    layer_config = self.base_layer_config.copy()
    layer_config['activation'] = 'linear'
    layer_config['name'] = self._name + '_' + name
    layer_config['kernel_initializer'] = tf.keras.initializers.get(kernel_initializer)
    return self.layer_class.from_config(layer_config)

  def _create_gate(self, gate_name):
    return GatingUnit(
      layer=self._create_layer(gate_name+'_gtrans', self.gates_initializer, use_bias=True),
      activation=self.gating_activation,
      name=self._name + '_' + gate_name
    )

  def _create_all_gates(self, input_shape):
    if not self.use_squeeze_gating:
      self.g_bpre_bpos = self._create_gate(gate_name='gu1_bpre_bpos')
      self.g_bn_ln = self._create_gate(gate_name='gu2_bn_ln')
      self.g_norm_non = self._create_gate(gate_name='gu3_norm_non')
      self.g_residual = self._create_gate(gate_name='gu4_res_or_not')
      self.g_proc_skip = self._create_gate(gate_name='gu5_proc_skip')
    else:
      self.g_bpre_bpos = self._create_squeeze_gate(
        input_shape=input_shape, gate_name='gu1_bpre_bpos',
        squeeze_type=self.use_squeeze_gating,
      )
      self.g_bn_ln = self._create_squeeze_gate(
        input_shape=input_shape, gate_name='gu2_bn_ln',
        squeeze_type=self.use_squeeze_gating,
      )
      self.g_norm_non = self._create_squeeze_gate(
        input_shape=input_shape, gate_name='gu3_norm_non',
        squeeze_type=self.use_squeeze_gating,
      )
      self.g_residual = self._create_squeeze_gate(
        input_shape=input_shape, gate_name='gu4_res_or_not',
        squeeze_type=self.use_squeeze_gating,
      )
      self.g_proc_skip = self._create_squeeze_gate(
        input_shape=input_shape, gate_name='gu5_proc_skip',
        squeeze_type=self.use_squeeze_gating,
      )
    return

  def _infer_output_shape(self, input_shape):
    keys = ['units', 'filters']

    for k in keys:
      if self.base_layer_config.get(k, None) is not None:
        return self.base_layer_config.get(k)

    return input_shape[-1]

  def compute_output_shape(self, input_shape):
    output_shape = self.layer.compute_output_shape(input_shape)
    return output_shape

  def _create_squeeze_gate(self, input_shape, gate_name, squeeze_type):
    output_units = self._infer_output_shape(input_shape)
    base_layer_activation = self.activation

    return SqueezeGatingUnit(
      input_shape=input_shape,
      output_units=output_units,
      base_layer_activation=base_layer_activation,
      activation=self.gating_activation,
      squeeze_type=squeeze_type,
      name=self._name + '_' + gate_name
    )

  def _create_layers(self):
    self.layer = self._original_layer.__class__.from_config(self.base_layer_config)
    assert self.layer.activation.__name__ == 'linear'
    
    # tf.keras.activations.get(self.activation)
    self.act_pre_bn = tf.keras.layers.Activation(
      self.activation, 
      name=self._name + '_{}_pre_bn'.format(self.activation))

    self.act_post_bn = tf.keras.layers.Activation(
      self.activation, 
      name=self._name + '_{}_post_bn'.format(self.activation))
    
    # bypass
    self.bypass = self._create_layer(
      'linear_bypass', 
      self.bypass_initializer, 
      use_bias=False
      )
    
    self.bn_pre_act = tf.keras.layers.BatchNormalization(name=self._name + '_bn_pre_act')
    self.bn_pos_act = tf.keras.layers.BatchNormalization(name=self._name + '_bn_post_act')    
    self.ln_pos_act = tf.keras.layers.LayerNormalization(name=self._name + '_ln_post_act')
    self.residual   = tf.keras.layers.Add(name=self._name + '_residual')
    return
  
  def _forward(self, inputs, for_model=False):
    tf_bypass = self.bypass(inputs)
    tf_x = self.layer(inputs)
    tf_x_act = self.act_pre_bn(tf_x)
    
    tf_x_bn_act = self.act_post_bn(self.bn_pre_act(tf_x))
    tf_x_act_bn = self.bn_pos_act(tf_x_act)
    tf_x_act_ln = self.ln_pos_act(tf_x_act)
    
    if not for_model:    
      tf_bpre_bpos = self.g_bpre_bpos([inputs, tf_x_bn_act, tf_x_act_bn])
      tf_bn_ln = self.g_bn_ln([inputs, tf_bpre_bpos, tf_x_act_ln])
      tf_norm_non = self.g_norm_non([inputs, tf_bn_ln, tf_x_act])
      tf_processed = tf_norm_non

      if self.add_residual_connection:
        tf_processed_res = tf_processed + tf_bypass
        tf_final_processed = self.g_residual([inputs, tf_processed, tf_processed_res])
      else:
        tf_final_processed = tf_processed

      tf_proc_noproc = self.g_proc_skip([inputs, tf_bypass, tf_final_processed])
    else:
      tf_bpre_bpos  = self.g_bpre_bpos._forward(
        inputs=[inputs, tf_x_bn_act, tf_x_act_bn],
        for_model=True)      
      tf_bn_ln = self.g_bn_ln._forward(
        inputs=[inputs, tf_bpre_bpos, tf_x_act_ln],
        for_model=True)
      tf_norm_non = self.g_norm_non._forward(
        inputs=[inputs, tf_bn_ln, tf_x_act],
        for_model=True)
      tf_processed = tf_norm_non

      if self.add_residual_connection:
        tf_processed_res = self.residual([tf_processed, tf_bypass])
        tf_final_processed = self.g_residual._forward(
          inputs=[inputs, tf_processed, tf_processed_res],
          for_model=True)
      else:
        tf_final_processed = tf_processed

      tf_proc_noproc = self.g_proc_skip._forward(
        inputs=[inputs, tf_bypass, tf_final_processed],
        for_model=True)      
    
    tf_out = tf_proc_noproc
    return tf_out      

  
  def get_model(self):
    inp_shape = get_input_shape(self.bypass)
    inp_name = "INPUT_" + self.name
    tf_inp = tf.keras.layers.Input(inp_shape, name=inp_name)
    tf_x = self._forward(tf_inp, for_model=True)
    return tf.keras.models.Model(tf_inp, tf_x, name=self.name + '_Model')
    

  def explain(self, log, np_input, save_fig=True):
    import os
    def P(s, color='y'):
      if log is not None:
        log.P(str(s), color=color)
      else:
        print("\x1b[1;33m" + str(s) + "\x1b[0m")
      return
    _layers = {
      self.g_bpre_bpos : 
        {
          'desc'  : "Gate1 - BN pre or post activation",
          'left'  : 'BN-pre-{}'.format(self.activation),
          'right' : 'BN-post-{}'.format(self.activation),
        },
      self.g_bn_ln : 
        {
          'desc'  : "Gate2 - BN or LayerNorm",
          'left'  : 'BN',
          'right' : 'LayerNorm',
        },
      self.g_norm_non :
        {
          'desc'  : "Gate3 - any norm or no norming at all",
          'left'  : 'Norming',
          'right' : 'No-norming',
        },
      self.g_proc_skip : 
        {
          'desc'  : "Gate5 - direct linear bypass or processed",
          'left'  : 'Bypass',
          'right' : 'Processed',
        },
      }

    if self.add_residual_connection:
      _layers[self.g_residual] = {
          'desc'  : 'Gate4 - residual or not',
          'left'  : 'No-residual',
          'right' : 'Residual',
      }

    input_shape = get_input_shape(self.layer)
    assert np_input.shape[1:] == input_shape, "np_input {} must match shape of MGU input {}".format(
      np_input.shape[:1], input_shape)
    P("  Layer: {} self-explainability analysis based on {} input".format(
      self.name,
      input_shape))
    activations = []
    for i, gate in enumerate(_layers):
      _desc = _layers[gate]['desc']
      _left = _layers[gate]['left']
      _right = _layers[gate]['right']
      _desc_formula = 'gate * ' + _left + "  +  (1 - gate) * "+_right
      P("    Analysing '{}'".format(
        _desc, 
        ))
      P("      Gate rule: `{}`".format(
        _desc_formula))
      gate_model = gate._get_gate_response_model()
      np_gate_activ = gate_model.predict(np_input)
      _gate_mean = np.mean(np_gate_activ)
      P("      Gate mean: {:.2f}, median: {:.2f}, min/max: {:.2f}/{:.2f}".format(
        _gate_mean,
        np.median(np_gate_activ),
        np.min(np_gate_activ),
        np.max(np_gate_activ),
        ))
      P("      Gate opened for: {}".format(
        _left if _gate_mean > 0.5 else _right
        ), color='g')
      activations.append(np_gate_activ.ravel())
    bins = 30
    _labels = [x['desc'] + '\n Low activations:'+ x['right']+ '; High activations:' + x['left'] 
               for x in list(_layers.values())]
    fn = os.path.join(log.get_output_folder(), self.name)
    log.plot_histogram(
      activations, 
      figsize=(13,8), 
      bins=bins,
      colors=['b','r','c','g'],
      labels=_labels,
      logscale=True,
      save_img_path=None if not save_fig else fn,
      xticks=[np.linspace(0,1, bins).round(2) for _ in range(len(_layers))],
      vline=0.5,
      )      
    return
    
            



import numpy as np
from time import time as tm

class _VectorSpaceMixin(object):
  """
  Mixin for vector space functionalities that are attached to `libraries.logger.Logger`:
    - embeddings
    - co-occurence matrix

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_MatplotlibMixin`:
    - self.add_copyright_to_plot
    - self.save_plot
  """

  def __init__(self):
    super(_VectorSpaceMixin, self).__init__()

    try:
      from .matplotlib_mixin import _MatplotlibMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _VectorSpaceMixin without having _MatplotlibMixin")

    return

  @staticmethod
  def _calc_delta(np_vsm):
    """
    method to calculate vector space model smoothing delta matrix. the input vsm can be a mco
    or even a non-square matrix
    """
    col_totals = np.array(np_vsm).sum(axis=0)
    row_totals = np.array(np_vsm).sum(axis=1)
    cm = [col_totals for _ in range(np_vsm.shape[0])]
    col_mat = np.vstack(cm)
    row_mat = row_totals.reshape((-1, 1))
    rm = [row_mat for _ in range(np_vsm.shape[1])]
    row_mat = np.hstack(rm)
    d1 = np_vsm / (np_vsm + 1)
    mins = np.minimum(col_mat, row_mat)
    d2 = mins / (mins + 1)
    delta = d1 * d2
    return delta

  @staticmethod
  def _csr_calc_delta(np_vsm):
    """
    method to calculate vector space model smoothing delta matrix. the input vsm can be a mco
    or even a non-square matrix
    """
    col_totals = np.array(np_vsm).sum(axis=0)
    row_totals = np.array(np_vsm).sum(axis=1)
    cm = [col_totals for _ in range(np_vsm.shape[0])]
    col_mat = np.vstack(cm)
    row_mat = row_totals.reshape((-1, 1))
    rm = [row_mat for _ in range(np_vsm.shape[1])]
    row_mat = np.hstack(rm)
    d1 = np_vsm / (np_vsm + 1)
    mins = np.minimum(col_mat, row_mat)
    d2 = mins / (mins + 1)
    delta = d1 * d2
    return delta

  @staticmethod
  def pmi(np_mco, positive=True, delta='raw'):
    """
    This function computes the (positive) Pointwise Mutual Information for a matrix of
    products co-occurence and includes the discounting approach

    np_mco :  np.ndarray
      The MCO matrix

    positive :  bool optional (True)
      all values will be positive (better usually)

    delta : str optional (default='raw')
      if delta is not None then the PMI will be discounted. 'pmi' will calculate
      the discount uwing the PMI and 'raw' will calculate based on MCO


    The output is a matrix where M[x,y] indicates the degree x and y co-occure together

    TODO:
      - percentual relationship strength

    """
    assert type(np_mco) in [np.ndarray]
    assert delta in [None, 'raw', 'pmi']

    str_msg = "Computing {}{}PMI on {}".format(
      "discounted ({}) ".format(delta) if delta is not None else "",
      "positive " if positive else "",
      np_mco.shape)
    np_col_totals = np_mco.sum(axis=0)
    total = np_col_totals.sum()
    np_row_totals = np_mco.sum(axis=1)
    np_expected = np.outer(np_row_totals, np_col_totals) / total
    np_oe = np_mco / np_expected
    with np.errstate(divide='ignore'):
      np_pmi = np.log(np_oe)
    np_pmi[np.isinf(np_pmi)] = 0.0
    if positive:
      np_pmi[np_pmi < 0] = 0.0
    if delta is not None:
      delta = _VectorSpaceMixin._csr_calc_delta(np_pmi if delta == 'pmi' else np_mco)
      np_pmi = np_pmi * delta
    return np_pmi.astype(np.float32)

  @staticmethod
  def csr_pmi2(csr_mco, positive=True, delta=None, batch_size=1000, DEBUG_STOP_AFTER_ITERS=0):
    """
    This function computes the (positive) Pointwise Mutual Information for a matrix of
    products co-occurence and includes the discounting approach

    np_mco :  np.ndarray
      The MCO matrix

    positive :  bool optional (True)
      all values will be positive (better usually)

    delta : str optional (default='pmi')
      if delta is not None then the PMI will be discounted. 'pmi' will calculate
      the discount uwing the PMI and 'raw' will calculate based on MCO


    The output is a matrix where M[x,y] indicates the degree x and y co-occure together

    TODO:
      - percentual relationship strength

    """
    from scipy import sparse
    assert type(csr_mco) in [sparse.csr.csr_matrix]
    assert delta in [None, 'raw', 'pmi']

    str_msg = "Computing {}{}PMI on sparse matrix {}".format(
      "discounted ({}) ".format(delta) if delta is not None else "",
      "positive " if positive else "",
      csr_mco.shape)

    np_col_totals = np.array(csr_mco.sum(axis=0)).flatten()
    np_row_totals = np.array(csr_mco.sum(axis=1)).flatten()
    total = np_col_totals.sum()
    csr_pmi = None
    batch_start = 0
    itrs = 0
    n_rows = csr_mco.shape[0]
    timings = []
    print("\r  " + str_msg + "...", flush=True, end='')
    while batch_start < n_rows:
      t_start = tm()
      batch_end = min(batch_start + batch_size, n_rows)
      np_batch_expected = np.outer(np_row_totals[batch_start:batch_end], np_col_totals) / total
      np_batch_aoe = csr_mco[batch_start:batch_end].toarray() / np_batch_expected
      np_batch_pmi = np.ma.log(np_batch_aoe).filled(0)
      if positive:
        np_batch_pmi[np_batch_pmi < 0] = 0.0
      csr_batch_pmi = sparse.csr_matrix(np_batch_pmi, dtype='float32')
      csr_pmi = sparse.vstack([csr_pmi, csr_batch_pmi], format='csr')
      t_end = tm()
      t_per_100 = (t_end - t_start) / (batch_end - batch_start) * 100
      timings.append(t_per_100)
      print("\r  " + str_msg + " {:.1f}%  - {:.3f} sec/ 100x rows".format(
        batch_end / n_rows * 100, np.mean(timings)), flush=True, end='')
      batch_start = batch_end
      itrs += 1
      if 0 < DEBUG_STOP_AFTER_ITERS <= itrs:
        break
    # end while
    if delta is not None:
      csr_delta = _VectorSpaceMixin._csr_calc_delta(csr_pmi if delta == 'pmi' else csr_mco)
      csr_pmi = csr_pmi * csr_delta
    return csr_pmi

  @staticmethod
  def csr_pmi(csr_mco, positive=True, delta=None, batch_size=2000, DEBUG_STOP_AFTER_ITERS=0):
    """
    This function computes the (positive) Pointwise Mutual Information for a matrix of
    products co-occurence and includes the discounting approach

    np_mco :  np.ndarray
      The MCO matrix

    positive :  bool optional (True)
      all values will be positive (better usually)

    delta : str optional (default='pmi')
      if delta is not None then the PMI will be discounted. 'pmi' will calculate
      the discount uwing the PMI and 'raw' will calculate based on MCO


    The output is a matrix where M[x,y] indicates the degree x and y co-occure together

    TODO:
      - percentual relationship strength

    """
    from scipy import sparse
    assert type(csr_mco) in [sparse.csr.csr_matrix]
    assert delta in [None, 'raw', 'pmi']

    str_msg = "Computing {}{}PMI on sparse matrix {}".format(
      "discounted ({}) ".format(delta) if delta is not None else "",
      "positive " if positive else "",
      csr_mco.shape)

    np_col_totals = np.array(csr_mco.sum(axis=0)).flatten()
    np_row_totals = np.array(csr_mco.sum(axis=1)).flatten()
    total = np_col_totals.sum()
    csr_pmi = None
    batch_start = 0
    itrs = 0
    n_rows = csr_mco.shape[0]
    timings = []
    print("\r  " + str_msg + "...", flush=True, end='')
    while batch_start < n_rows:
      t_start = tm()
      batch_end = min(batch_start + batch_size, n_rows)
      np_mco_batch = csr_mco[batch_start:batch_end].toarray()
      np_batch_expected = (np.outer(np_row_totals[batch_start:batch_end], np_col_totals) / total).astype(np.float32)
      np_batch_aoe = (np_mco_batch / np_batch_expected).astype(np.float32)
      csr_batch_pmi = sparse.csr_matrix(np_batch_aoe, dtype='float32')
      csr_batch_pmi.data = np.log(csr_batch_pmi.data)
      if positive:
        csr_batch_pmi[csr_batch_pmi < 0] = 0.0
      csr_pmi = sparse.vstack([csr_pmi, csr_batch_pmi], format='csr')
      t_end = tm()
      t_per_100 = (t_end - t_start) / (batch_end - batch_start) * 100
      timings.append(t_per_100)
      print("\r  " + str_msg + " {:.1f}%  - {:.3f} sec/ 100x rows".format(
        batch_end / n_rows * 100, np.mean(timings)), flush=True, end='')
      batch_start = batch_end
      itrs += 1
      if 0 < DEBUG_STOP_AFTER_ITERS <= itrs:
        break
    # end while
    if delta is not None:
      csr_delta = _VectorSpaceMixin._csr_calc_delta(csr_pmi if delta == 'pmi' else csr_mco)
      csr_pmi = csr_pmi * csr_delta
    return csr_pmi

  def get_embeds_map(self, embed_matrix, figsize=None, labels=None, colors=None, plot=True,
                     title='Embedding map', return_fig=False):
    from sklearn.manifold import TSNE
    import matplotlib.pyplot as plt
    assert len(embed_matrix.shape) == 2
    nr_embs = embed_matrix.shape[0]
    if nr_embs > 1000:
      self.P("Number of embeddings quite large. Map calculcation will take some time...")
    if labels is not None:
      assert len(labels) == nr_embs, "Number of labels {} must equal number of obs in emb space {}".format(
        len(labels), nr_embs)
    self.P("Calculating map for embeds {}...".format(embed_matrix.shape))
    tsne = TSNE(verbose=2)
    tsne.fit(embed_matrix)
    self.P(" Done calculating map for embeds {}.".format(embed_matrix.shape), show_time=True)
    if plot:
      x = tsne.embedding_[:, 0]
      y = tsne.embedding_[:, 1]
      fig, ax = plt.subplots(figsize=figsize)
      ax.scatter(x, y, c=colors)
      if labels is not None:
        for i, label in enumerate(labels):
          ax.annotate(label, (x[i], y[i]))
      plt.title(title)
      self.add_copyright_to_plot(plt)
      plt.show()
      self.save_plot(plt, label=title)
    if return_fig:
      return tsne.embedding_, plt
    else:
      return tsne.embedding_

  @staticmethod
  def _measure_changes(Y, Y_prev):
    """
    this helper function measures changes in the embedding matrix
    between previously step of the retrofiting loop and the current one
    """
    return np.abs(np.mean(np.linalg.norm(
      np.squeeze(Y_prev) - np.squeeze(Y),
      ord=2
    )))


import json
import traceback

from datetime import datetime
from ratio1.logging.logger_mixins.datetime_mixin import _DateTimeMixin as dt_utils


def get_config_from_code(fn, log=None):
  if log is None:
    class _L:
      def P(self, *args, **kwargs):
        print(args[0])
    log = _L()
  result = "{"
  try:
    with open(fn,'rt') as fh:
      str_code = fh.read()
  except:
    return None
  start_pos = str_code.index("_CONFIG")
  str_to_parse = str_code[start_pos:]
  start_pos = str_to_parse.index("{")
  str_to_parse = str_to_parse[start_pos+1:]
  opened = 1
  cnt = 0
  while opened:
    ch = str_to_parse[cnt]
    if ch == '{':
      opened += 1
    elif ch == '}':
      opened -= 1
    result = result + ch    
    cnt += 1
  try:
    dct_result = eval(result)
  except:
    str_err = traceback.format_exc()
    log.P("Error deconding _CONFIG data from {}:\n{}".format(fn, str_err), color='r')
    return None
  return dct_result


def get_now_value_from_time_dict(dct_hours):
  conf_thr = None
  for hours in dct_hours:
    t1, t2 = hours.split('-')
    if dt_utils.now_in_interval_hours(start=t1, end=t2):
      conf_thr = dct_hours[hours]
      break
    # endif dt_utils.now_in_interval_hours(start=t1, end=t2)
  # endfor hours in dct_hours
  return conf_thr
  
if __name__ == '__main__':
  d ={
    "10:00-17:00" : 0.3,
    "17:00-10:00" : 0.5,
    }
  
  print(get_now_value_from_time_dict(d)) ### TODO L this returns None!!
  
  loc1 = [
         {
           'NAME' : 'TOP',
           'VALUE' : 100,
         },
         {
           'NAME' : 'LEFT',
           'VALUE' : 100,
         },
         {
           'NAME' : 'INSTANCE_ID',
           'VALUE' : 'TEST1',
         },
      ]
  
  loc2 = {
    # 'BLABLA1' : 1,
    # 'BLABLA2' : 2,
    # 'DESCRIPTION' : "12312312321",
    
    'PLUGIN_INSTANCE_PARAMETER_LIST' : [
         {
           'NAME' : 'TOP',
           'VALUE' : 100,
         },
         {
           'NAME' : 'LEFT',
           'VALUE' : 100,
         },
         {
           'NAME' : 'INSTANCE_ID',
           'VALUE' : 'TEST1',
         },         
         {
           'NAME'  : 'POINTS',
           'VALUE' :  '[[100,100], [0,0]]',
           'TYPE'  : 'LIST'
         }
      ]    
    }

  from plugins.io_formatters.cavi2 import process_instance_config
  print(json.dumps(process_instance_config(loc2), indent=4))
  



"""
[Y8][23-03-31 20:50:43] Timing results at 2023-03-31 20:50:43:
[Y8][23-03-31 20:50:43] Section 'main' last seen 0.0s ago
[Y8][23-03-31 20:50:43]  y5l6 = 0.0429s/q:0.0429s/nz:0.0429s, max: 0.0468s, lst: 0.0429s, c: 100/L:6%
[Y8][23-03-31 20:50:43]    y5l6_nms = 0.0245s/q:0.0245s/nz:0.0245s, max: 0.0296s, lst: 0.0270s, c: 100/L:14%
[Y8][23-03-31 20:50:43]      y5l6_nms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0004s, lst: 0.0001s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y5l6_inclnms = 0.0433s/q:0.0433s/nz:0.0433s, max: 0.0471s, lst: 0.0426s, c: 100/L:9%
[Y8][23-03-31 20:50:43]    y5l6_inclnms_cpu = 0.0002s/q:0.0002s/nz:0.0199s, max: 0.0006s, lst: 0.0001s, c: 100/L:99%
[Y8][23-03-31 20:50:43]  y5s6 = 0.0126s/q:0.0126s/nz:0.0126s, max: 0.0307s, lst: 0.0283s, c: 100/L:0%
[Y8][23-03-31 20:50:43]    y5s6_nms = 0.0017s/q:0.0017s/nz:0.0017s, max: 0.0037s, lst: 0.0023s, c: 100/L:0%
[Y8][23-03-31 20:50:43]      y5s6_nms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0004s, lst: 0.0002s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y5s6_inclnms = 0.0123s/q:0.0123s/nz:0.0123s, max: 0.0271s, lst: 0.0234s, c: 100/L:0%
[Y8][23-03-31 20:50:43]    y5s6_inclnms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0003s, lst: 0.0002s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y8l = 0.0329s/q:0.0329s/nz:0.0329s, max: 0.0356s, lst: 0.0326s, c: 100/L:11%
[Y8][23-03-31 20:50:43]    y8l_nms = 0.0187s/q:0.0187s/nz:0.0187s, max: 0.0223s, lst: 0.0203s, c: 100/L:12%
[Y8][23-03-31 20:50:43]      y8l_nms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0004s, lst: 0.0002s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y8l_inclnms = 0.0329s/q:0.0329s/nz:0.0329s, max: 0.0371s, lst: 0.0329s, c: 100/L:4%
[Y8][23-03-31 20:50:43]    y8l_inclnms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0005s, lst: 0.0002s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y8s = 0.0110s/q:0.0110s/nz:0.0110s, max: 0.0268s, lst: 0.0094s, c: 100/L:0%
[Y8][23-03-31 20:50:43]    y8s_nms = 0.0018s/q:0.0018s/nz:0.0018s, max: 0.0035s, lst: 0.0016s, c: 100/L:0%
[Y8][23-03-31 20:50:43]      y8s_nms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0003s, lst: 0.0002s, c: 100/L:100%
[Y8][23-03-31 20:50:43]  y8s_inclnms = 0.0100s/q:0.0100s/nz:0.0100s, max: 0.0254s, lst: 0.0089s, c: 100/L:0%
[Y8][23-03-31 20:50:43]    y8s_inclnms_cpu = 0.0002s/q:0.0002s/nz:-1.0000s, max: 0.0003s, lst: 0.0001s, c: 100/L:100%
[Y8][23-03-31 20:50:43] Section 'LOGGER_internal' last seen 0.0s ago
[Y8][23-03-31 20:50:43]  _logger = 0.0030s/q:0.0032s/nz:0.0032s, max: 0.0047s, lst: 0.0020s, c: 122/L:17%
[Y8][23-03-31 20:50:43]    _logger_add_log = 0.0010s/q:0.0011s/nz:0.0011s, max: 0.0019s, lst: 0.0006s, c: 122/L:18%
[Y8][23-03-31 20:50:43]    _logger_save_log = 0.0019s/q:0.0021s/nz:0.0021s, max: 0.0035s, lst: 0.0014s, c: 122/L:17%
[Y8][23-03-31 20:50:43] Model y5l6 gain -0.00047s, equal: True
[Y8][23-03-31 20:50:43] Model y5s6 gain 0.00031s, equal: True
[Y8][23-03-31 20:50:43] Model y8l gain -0.00007s, equal: True
[Y8][23-03-31 20:50:43] Model y8s gain 0.00097s, equal: True


pip install ultralytics[export]


TODO:
    - finish comparision nms y5 vs y8
    - add nms in y5 DONE
    - add nms in y8 DONE
    - optimize nms5
    - optimize nms8
    - resave, load and send to testing!
    
"""
import numpy as np
import os
import sys
import torch as th
import torchvision as tv
import json
import cv2

import time
import argparse
import shutil
import gc
from pathlib import Path
import itertools as it

try:
  from ultralytics import YOLO
except:
  print("WARNING: Only TESTS and GENERATE_FULL will work !", flush=True)

from naeural_core import Logger
from decentra_vision.draw_utils import DrawUtils
from naeural_core.local_libraries.nn.th.utils import th_resize_with_pad
from plugins.serving.architectures.y5.general import scale_coords

from naeural_core.xperimental.th_y8.utils import predict
from naeural_core.xperimental.th_y8.utils import Y5, Y8, BackendType
  
from naeural_core.xperimental.onnx.utils import create_from_torch
from naeural_core.xperimental.th_y8.y_generate import yolo_models

def get_test_images():
  folder = os.path.split(__file__)[0]
  img_names = [
    'bus.jpg',
    'faces9.jpg',
    'faces21.jpg',
    'img.png',
    '688220.png',
    'bmw_man3.png',
    'faces3.jpg',
    'LP3.jpg',
    'pic1_crop.jpg'
  ]
  models_for_nms_prep = [
    {
      'pts': r'C:\repos\ultralytics\__personal\traces\y8x_1152x2048.torchscript',
      'model': 'y8x_1152x2048'
    }
  ]

  # assert img is not None
  imgs = [
    cv2.imread(os.path.join(folder, im_name))
    for im_name in img_names
  ]
  return imgs


def get_shape_for_pt_file(filename):
  sizes = filename.split('_')
  sizes = sizes[2]
  sizes = sizes.split('x')
  return (int(sizes[0]), int(sizes[1]))

def load_ths(path, log, dev, half=None, max_batch_size=None):
  extra_files = {'config.txt' : ''}
  loaded_model = th.jit.load(
    path,
    map_location=dev,
    _extra_files=extra_files
  )
  config = json.loads(extra_files['config.txt' ].decode('utf-8'))
  if half is not None and half:
    loaded_model = loaded_model.half()
  return loaded_model, config

def load_trt(path, log, dev, half=None, max_batch_size=None):
  from naeural_core.serving.base.backends.trt import TensorRTModel
  loaded_model = TensorRTModel(log=log)
  if half is None and max_batch_size is None:
    engine_p, config_p = TensorRTModel._get_engine_and_config_path(path, half, max_batch_size)

    loaded_model.load_model(engine_p, config_p, device=dev)
    max_batch_size = loaded_model._metadata[TensorRTModel.METADATA_MAX_BATCH]
    half = (loaded_model._metadata[TensorRTModel.ONNX_METADATA_KEY]['precision'] == 'fp16')
    loaded_model.check_engine_file(
      dev,
      path,
      config_p,
      engine_p,
      half,
      max_batch_size
    )
  else:
    loaded_model.load_or_rebuild_model(
      onnx_path=path,
      fp16=half,
      max_batch_size=max_batch_size,
      device=dev
  )
  config = loaded_model._metadata[TensorRTModel.ONNX_METADATA_KEY]
  return loaded_model, config

def load_onnx(path, log, dev, half=None, max_batch_size=None):
  from naeural_core.serving.base.backends.onnx import ONNXModel
  if max_batch_size is None:
    max_batch_size = 64
  loaded_model = ONNXModel()
  loaded_model.load_model(
    model_path=path,
    max_batch_size=max_batch_size,
    device=dev
  )
  config = loaded_model.get_metadata()
  return loaded_model, config

def load_model(path, log, dev, half=None, max_batch_size=None):
  if path.endswith('.ths') or path.endswith('.torchscript'):
    return load_ths(path, log, dev, half, max_batch_size)
  elif path.endswith('_trt.onnx'):
    return load_trt(path, log, dev, half, max_batch_size)
  elif path.endswith('.onnx'):
    return load_onnx(path, log, dev, half, max_batch_size)
  raise ValueError("Unknown model extension")

def is_valid_model_input(filename):
  if ('_nms' not in file_name) and ('y' in file_name):
    return False

def get_model_config_name(file_name):
  ext_pos = file_name.rfind('.')
  if ext_pos == -1:
    # File doesn't have extension, no idea what this is so skip
    return None

  extension = file_name[ext_pos:]
  if extension not in ['.ths', '.torchscript']:
    return None

  # Get the filename without the extension.
  fn = file_name[:ext_pos]

  if ('_nms' in file_name) or ('y' not in fn):
    return None

  if 'lpd' in fn:
    fn = fn[fn.find('lpd'):]
  else:
    fn = fn[fn.find('y'):]
  return fn

def is_valid_for_export(precision, device_str, topk, backend):
  # TODO: maybe make all traces on precision 16 invalid because we can
  # always cast them to 16 at the model loading.
  if device_str == 'cpu' and precision == 16:
    # FP16 will crash and burn on cpu
    return False
  # device 'cpu' and backend 'ths' is not valid because the
  # 'ths' trace on gpu should work on cpu too
  if device_str == 'cpu' and backend != 'onnx':
    return False
  if device_str == 'cuda' and backend not in ['ths', 'trt']:
    return False
  if topk and backend != 'ths':
    # Only torch support topk at the moment
    return False
  if precision == 16 and backend not in ['onnx', 'trt']:
    # We only need fp16 binaries when exporting to onnx
    return False
  return True


def get_generation_configs(log):
  models_folder = log.get_models_folder()
  files = os.listdir(models_folder)
  # Maybe filter only yolo models or have only yolo models in the folder
  export_grid = {
    'topk' : [True, False],
    'device' : ['cpu', 'cuda'],
    'precision' : [16, 32]
  }

  combinations = list(it.product(*(export_grid[Param] for Param in export_grid.keys())))

  for file_name in files:
    fn = get_model_config_name(file_name)
    if fn is None:
      continue

    for topk, device, precision in combinations:
      if is_valid_for_export(precision, device, topk, 'ths'):
        continue
        conf = {
          'path' : os.path.join(models_folder, file_name),
          'model_name': fn,
          'backend' : 'ths',
          'precision' : precision,
          'device' : device,
          'topk' : topk
        }
        yield conf
    #endfor all combinations
  #endfor all models

  # For onnx/trt we need to start from pytorch model directly, so the exported
  # torchscript files won't help us (except for getting the actual configuration).
  for model, export_kwargs in yolo_models(log):
    pt_path = os.path.join(models_folder, model.model.pt_path)
    extra_files = {'config.txt' : ''}
    fn = get_model_config_name(pt_path)

    # Read the config from the already exported model.
    # Ideally YOLO would give us a way to generate this ourselves.
    ths_m = th.jit.load(pt_path, map_location='cpu', _extra_files=extra_files)
    config = json.loads(extra_files['config.txt' ].decode('utf-8'))
    shape = config['imgsz']
    for backend in ['trt', 'onnx']:
      for topk, device, precision in combinations:
        if is_valid_for_export(precision, device, topk, backend):
          real_model = model.model
          if precision == 16:
            real_model = real_model.half()
          conf = {
            'path' : os.path.join(models_folder, pt_path),
            'config' : config,
            'backend' : backend,
            'model_name' : fn,
            'th_model' : real_model,
            'precision' : precision,
            'device' : device,
            'topk' : topk
          }
          yield conf
  #endfor
  return

def get_test_configs(log):
  models_folder = log.get_models_folder()
  files = os.listdir(models_folder)

  test_grid = {
    'device' : ['cpu', 'cuda'],
    'precision' : [16, 32]
  }
  combinations = list(it.product(*(test_grid[Param] for Param in test_grid.keys())))

  for file_name in files:
    backend = None
    is_onnx = file_name.endswith('.onnx')
    is_trt = file_name.endswith('_trt.onnx')
    is_ths = file_name.endswith('.ths') or file_name.endswith('.torchscript')
    if is_onnx and not is_trt:
      backend = 'onnx'
    if is_trt:
      backend = 'trt'
    if is_ths:
      backend = 'ths'
    if backend is None:
      continue

    is_half = 'f16' in file_name
    for device, precision in combinations:
      if (is_onnx or is_trt) and is_half != (precision==16):
        continue
      if is_trt and device != 'cuda':
        continue
      if (is_onnx and not is_trt) and device != 'cpu':
        continue
      if precision == 16 and device == 'cpu':
        continue
      if is_half and precision != 16:
        continue
      yield {
        'path': os.path.join(models_folder, file_name),
        'model_name': file_name,
        'precision': precision,
        'backend': backend,
        'device': device
      }
    #endfor
  #endfor
  return

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description='Build flavours of YOLO')
  parser.add_argument(
    "--notest",
    action='store_true',
    help='Skip running validation tests'
  )
  parser.add_argument(
    "--show",
    action='store_true',
    help='Show image results'
  )
  args = parser.parse_args()

  GENERATE_FULL = True
  GENERATE_TOPK = True
  TOP_K_VALUES = [False, True] if GENERATE_TOPK else [False]
  generated_models = []
  TRANSFER_CLASSNAMES = False
  CLASSNAMES = None
  CLASSNAMES_DONATE_FROM = os.path.join('_local_cache', '_models', '20230723_y8l_nms.ths') 
  TEST = not args.notest
  SHOW = args.show
  SHOW_LABELS = True
  dev = th.device('cuda')

  top_candidates = 6

  if not th.cuda.is_available():
    log.P('CUDA is not available, exiting')
    exit(-1)
  if th.cuda.device_count() < 1:
    log.P('No available CUDA devices, exiting')
    exit(-1)

  if CLASSNAMES is None:
    extra_files = {'config.txt': ''}
    model = th.jit.load(CLASSNAMES_DONATE_FROM, map_location='cpu', _extra_files=extra_files)
    CLASSNAMES = json.loads(extra_files['config.txt'])['names']
  # end if CLASSNAMES is None

  log = Logger('Y8', base_folder='.', app_folder='_local_cache')
  models_folder = log.get_models_folder()
  model_date = time.strftime('%Y%m%d', time.localtime(time.time()))
  pyver = sys.version.split()[0].replace('.','')
  thver = th.__version__.replace('.','')

  # assert img is not None
  origs = get_test_images()
  images = origs
  
  model_ext = {
    'ths' : 'ths',
    'trt' : 'onnx',
    'onnx' : 'onnx'
  }

  if GENERATE_FULL:
    for m in get_generation_configs(log):
      export_type = m['backend']
      trt = (export_type == 'trt')
      onnx = (export_type == 'onnx')
      fn_path = m['path']
      y_model = fn_path
      topk = m['topk']
      device = m['device']
      config = m.get('config')
      precision = m['precision']
      if device.startswith('cuda'):
        dev = th.device('cuda:0')
      else:
        dev = th.device(device)

      backend_type = BackendType.TORCH
      if trt and not device.startswith('cuda'):
        # Only the cuda device is supported for trt
        continue

      if trt or onnx:
        # We can't use torchscript in the model for trt so load it as a torch model.
        y_model_arg = (m['th_model'], m['config'])
        backend_type = BackendType.TENSORRT if trt else BackendType.ONNX
      else:
        y_model_arg = fn_path

      log.P("Generating for: {} and backend {}".format(m['path'], m['backend']))
      fn_path = m['path']
      model_name = os.path.splitext(os.path.split(fn_path)[1])[0]
      is_y5 = '_y5' in model_name

      if is_y5 and (onnx or trt):
        # No support for y5 in onnx at the moment (although
        # may be trivial to add).
        continue

      if is_y5:
        model = Y5(y_model_arg, dev=dev, topk=topk)
      else:
        model = Y8(y_model_arg, dev=dev, topk=topk, backend_type=backend_type)

      model.eval()
      # fn_model_name = model_name[model_name.find('y'):]
      # # Remove model extension
      # fn_model_name = fn_model_name[:fn_model_name.rfind('.')]
      fn_model_name = model_name[model_name.find('y'):].split('.ths')[0]
      imgsz = model.config.get('imgsz')
      if imgsz is None:
        imgsz = model.config.get('shape')

      input_names = ['inputs']
      output_names = ['preds', 'num_preds']
      config = {
        **model.config,
        'input_shape': (*imgsz, 3),
        'python': sys.version.split()[0],
        'torch': th.__version__,
        'torchvision': tv.__version__,
        'device': device,
        'optimize': False,
        'n_candidates': top_candidates,
        'date': model_date,
        'model': fn_model_name,
        'names': CLASSNAMES if TRANSFER_CLASSNAMES else model.config['names'],
        'input_names' : input_names,
        'output_names' : output_names,
        'includes_nms' : True,
        'includes_topk' : topk
      }
      config['includes_nms'] = True
      config['includes_topk'] = topk
      if export_type == 'trt' or export_type == 'onnx':
        # ONNX is strongly typed so whatever datatype we have is baked
        # into it. We therefore need to record the precision since we
        # can't cast the model on the target. This is not related to TRT.
        config['precision'] = 'fp' + str(precision)

      log.P("  Model {}: {},{}".format(model_name, fn_path, list(model.config.keys())), color='m')

      h, w = imgsz[-2:]
      log.P("  Resizing from {} to {}".format([x.shape for x in images],(h,w)))
      results = th_resize_with_pad(
        img=images,
        h=h,
        w=w,
        device=dev,
        normalize=True,
        half=(precision==16),
        return_original=False
      )
      if len(results) < 3:
        prep_inputs, lst_original_shapes = results
      else:
        prep_inputs, lst_original_shapes, lst_original_images = results

      model.to(dev)

      model_name_suffix = f'nms_top{top_candidates}' if topk else 'nms'
      dtype_suffix = '' if not (trt or onnx) else ('_f' + str(precision))
      # Add a backend info descriminator in the name to separate pure ONNX models
      # from tensorrt ones.
      backend_info = ''
      if trt:
        backend_info = '_trt'
      extension = model_ext[export_type]
      fn = os.path.join(
        models_folder,
        f'{model_date}_{fn_model_name}_{model_name_suffix}{dtype_suffix}{backend_info}.{extension}'
      )

      if export_type == 'ths':
        log.P("  Scripting...")
        traced_model = th.jit.trace(model, prep_inputs, strict=False)
        log.P("  Forwarding using traced...")
        output = traced_model(prep_inputs)
        extra_files = {'config.txt' : json.dumps(config)}
        log.P(f"  Saving '{fn}'...")
        traced_model.save(fn, _extra_files=extra_files)
      elif export_type == 'trt' or export_type == 'onnx':
        # This should be irrelevant. Left in case of error at testing
        # from naeural_core.serving.base.backends.trt import TensorRTModel
        if export_type == 'trt':
          log.P("  Exporting to ONNX for TensorRT...")
          log.P("  Exporting with config {}".format(config))
          log.P("  Saving '{}'...".format(fn))
        elif export_type == 'onnx':
          log.P("  Exporting to ONNX for ONNX/OpenVINO...")
          log.P("  Exporting with config {}".format(config))
          log.P("  Saving '{}'...".format(fn))
        create_from_torch(
          model,
          dev,
          fn,
          precision==16,
          config['input_names'],
          config['output_names'],
          args=prep_inputs,
          batch_axes=None,
          metadata=config
        )
        log.P("  Done")
      else:
        raise ValueError("Unknown backend {}".format())

      del model
      gc.collect()
      th.cuda.empty_cache()

      loaded_model, config = load_model(fn, log, dev, precision==16, 2*len(images))
      log.P("  Loaded config with {}".format(list(config.keys())))
      prep_inputs_test = th.cat([prep_inputs,prep_inputs])
      log.P("  Running forward...")
      res, n_det = loaded_model(prep_inputs_test)
      log.P("  Done running forward. Ouput:/n{}".format(res.shape))

      del loaded_model
      gc.collect()
      th.cuda.empty_cache()
    # endfor m in models
  # endif GENERATE_FULL

  dev = th.device('cuda')

  if TEST:
    painter = DrawUtils(log=log)
    dct_results = {}
    for m in get_test_configs(log):
      print("Running for {}".format(m))
      dev = th.device(m['device'])
      fn_path = m['path']
      use_fp16 = m['precision'] == 16
      model, config = load_model(
        fn_path,
        log,
        dev,
        half=use_fp16,
        max_batch_size=2*len(images)
      )
      if use_fp16 and m['backend'] == 'ths':
        model = model.half()
      imgsz = config.get('imgsz', config.get('shape'))
      includes_nms = config.get('includes_nms')
      includes_topk = config.get('includes_topk')
      model_name = m['model_name']
      model_result_name = model_name
      if m['backend'] == 'ths':
        model_result_name = (model_name + '_fp16') if use_fp16 else model_name
      if model_result_name not in dct_results:
        dct_results[model_result_name] = {}
      is_y5 = 'y5' in model_name
      if includes_topk:
        model_name = model_result_name + '_topk'
      else:
        model_name = model_result_name + ('_inclnms' if includes_nms else '')
      log.P("Model {}: {}, {}:".format(model_name, imgsz, fn_path,), color='m')
      maxl = max([len(k) for k in config])
      for k,v in config.items():
        if not (isinstance(v, dict) and len(v) > 5):
          log.P("  {}{}".format(k + ':' + " " * (maxl - len(k) + 1), v), color='m')

      h, w = imgsz[-2:]
      log.P("  Resizing from {} to {}".format([x.shape for x in images],(h,w)))
      class_names = config['names']
      results = th_resize_with_pad(
        img=images + images,
        h=h,
        w=w,
        device=dev,
        normalize=True,
        return_original=False,
        half=use_fp16
      )
      if len(results) < 3:
        prep_inputs, lst_original_shapes = results
      else:
        prep_inputs, lst_original_shapes, lst_original_images = results

      # warmup
      log.P("  Warming up...")
      for _ in range(20):
        print('.', flush=True, end='')
        pred_nms_cpu = predict(model, prep_inputs, model_name, config, log=log, timing=False)
      print('')

      N_TESTS = 100 if 'cuda' in m['device'].lower() else 10
      # timing
      log.P("  Predicting...")
      for _ in range(N_TESTS):
        print('.', flush=True, end='')
        pred_nms_cpu = predict(model, prep_inputs, model_name, config, log=log, timing=True)
      print('')

      log.P("  Last preds:\n{}".format(pred_nms_cpu))

      mkey = 'includes_topk' if includes_topk else 'includes_nms' if includes_nms else 'normal'

      dct_results[model_result_name][mkey] = {
          'res'  : pred_nms_cpu,
          'time' : log.get_timer_mean(model_name),
          'name' : model_name
      }

      if SHOW:
        log.P("  Showing...")
        for i in range(len(images)):
          # now we have each individual image and we generate all objects
          # what we need to do is to match `second_preds` to image id & then
          # match second clf with each box
          img_bgr = origs[i].copy()
          np_pred_nms_cpu = pred_nms_cpu[i]
          original_shape = lst_original_shapes[i]
          np_pred_nms_cpu[:, :4] = scale_coords(
            img1_shape=(h, w),
            coords=np_pred_nms_cpu[:, :4],
            img0_shape=original_shape,
          ).round()
          lst_inf = []
          for det in np_pred_nms_cpu:
            det = [float(x) for x in det]
            # order is [left, top, right, bottom, proba, class] => [L, T, R, B, P, C, RP1, RC1, RP2, RC2, RP3, RC3]
            L, T, R, B, P, C = det[:6]  # order is [left, top, right, bottom, proba, class]
            label = class_names[str(int(C))] if SHOW_LABELS else ''
            img_bgr = painter.draw_detection_box(image=img_bgr, top=int(T), left=int(L), bottom=int(B), right=int(R), label=label, prc=P)
          painter.show(fn_path, img_bgr, orig=(0, 0))
        #endfor plot images
      #endif show

      del model
      gc.collect()
      th.cuda.empty_cache()
    # endfor each model
  # endif test models
  log.show_timers()
  for mn in dct_results:
    if 'normal' not in dct_results[mn] or 'includes_nms' not in dct_results[mn]:
      continue
    a1 = dct_results[mn]['normal']['res']
    t1 = dct_results[mn]['normal']['time']
    a2 = dct_results[mn]['includes_nms']['res']
    t2 = dct_results[mn]['includes_nms']['time']
    # a3 = dct_results[mn]['includes_topk']['res']
    # t3 = dct_results[mn]['includes_topk']['time']
    # a3 = [x[:, :6] for x in a3]
    
    ok1 = all([np.allclose(a1[i], a2[i]) for i in range(len(a1))])
    # ok2 = all([np.allclose(a1[i], a3[i]) for i in range(len(a1))])

    gain1 = t1-t2
    rel_gain1 = gain1 / t1
    # gain2 = t1-t3
    log.P(f'Model with NMS {mn} {t1} => {t2}[gain {gain1:.5f}s({rel_gain1:.2f}%)], equal: {ok1}', color='r' if not ok1 else 'g')
    # log.P("Model with NMS {} gain {:.5f}s, equal: {}".format(mn, gain1, ok1), color='r' if not ok1 else 'g')
    # log.P("Model with TopK {} gain {:.5f}s, equal: {}".format(mn, gain2, ok2), color='r' if not ok2 else 'g')



import abc
from typing import Tuple
import torch as th

class ModelBackendWrapper(metaclass=abc.ABCMeta):
  """
  A model wrapper for inference, abstracting the backend
  that is actually used (TensorRT, Torch, ONNX, etc). Users
  should interact with the model only through this interface
  to ensure portability to new backends.
  """

  @abc.abstractmethod
  def get_device(self) -> th.device :
    """
    Returns the torch device used by the model.
    """
    raise NotImplementedError("get_device not implemented")

  @abc.abstractmethod
  def get_input_dtype(self, index : int) -> th.dtype:
    """
    Get the torch dtype for a given argument number.

    Parameters:
      - index : int, the number of the argument

    Returns:
      The torch dtype of the argument.
    """
    raise NotImplementedError("get_device not implemented")

  @abc.abstractmethod
  def __call__(self, *args : th.tensor) -> Tuple[th.tensor, ...]:
    """
    Performs inference by invoking the forward method of the model.

    Parameters
    - All inputs are unnamed torch tensor arguments specific to the model.

    Returns
    - Tuple of tensors (specific to the model) if the model has more than
      one ouput. The model itself does not capture these.
      If the model only has one output (a tensor), returns that output
      as is (a pytorch tensor).
    """
    raise NotImplementedError("__call__ not implemented")



import numpy as np

from naeural_core import Logger

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='EE_XPER', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )
  
  N = 14
  H = 1520
  W = 2688  
  ITERS = 100  
  for _ in range(ITERS):
    lst_imgs = [np.random.randint(0, 256, (H, W, 3), dtype=np.uint8) for _ in range(N)]
    log.start_timer('allocate')
    np_imgs = np.array(lst_imgs)
    log.stop_timer('allocate')
    
  np_cache = np.zeros_like(np_imgs, dtype=np.uint8)
  np_imgs = np.random.randint(0, 256, (N, H, W, 3), dtype=np.uint8)  
  for _ in range(ITERS):
    np_imgs-= 1
    log.start_timer('copy')
    np_cache[:,:] = 0
    for i in range(N):
      np_cache[i, :] = np_imgs[i]
    log.stop_timer('copy')
  
  log.show_timers()
  


import math
import os
import subprocess
import tempfile
from datetime import datetime, timedelta
from time import time


class FFMPEGUtils(object):
  def __init__(self, *, caller=None):
    self._caller = caller
    return

  def P(self, text, **kwargs):
    if self._caller is not None:
      self._caller.P(text, **kwargs)
    else:
      print(text)

  def _call_subprocess(self, command):
    uid = time()
    outf = 'output_{}.txt'.format(uid)
    errf = 'error_{}.txt'.format(uid)

    outfd = open(outf, 'xb')
    errfd = open(errf, 'xb')

    proc = subprocess.Popen(
      command,
      shell=True,
      stdout=outfd,
      stderr=errfd
    )
    proc.wait()

    outfd.close()
    errfd.close()

    outfd = open(outf, 'rb')
    errfd = open(errf, 'rb')

    stdout = outfd.read().decode()
    stderr = errfd.read().decode()

    outfd.close()
    errfd.close()

    os.remove(outf)
    os.remove(errf)

    self._process_output(proc.returncode, stdout=stdout, stderr=stderr, command=command)

    return stdout, stderr

  def video_file_duration(self, path, *, as_float=False):
    command = "ffprobe -i \"{}\"".format(path)

    _, stderr = self._call_subprocess(command)

    duration = stderr.split('Duration: ')[1].split(',')[0]

    if not as_float:
      return duration

    dt_duration = datetime.strptime(duration, '%H:%M:%S.%f')
    dt_zero = datetime.strptime('1900-01-01', '%Y-%m-%d')
    total_seconds = (dt_duration - dt_zero).total_seconds()
    return total_seconds

  def video_resolution(self, path):
    command = "ffprobe -v error -select_streams v:0 -show_entries stream=width,height -of csv=s=x:p=0 \"{}\"".format(
      path)

    stdout, _ = self._call_subprocess(command)

    width, height = stdout.split('\r\n')[0].split('x')[:2]

    return (int(height), int(width))

  def _process_output(self, returncode, *, stdout, stderr, command=None):
    if returncode != 0:
      self.P("Return code != 0, maybe an error!", color='r')
      if self._caller is not None:
        self._caller._create_notification(
          notif='EXCEPTION',
          msg="Return code not 0, maybe an error! The command was: {}\n The output of the process was: {}\n The stderr of the process was {}\n Raising exception...".format(
            command, stdout, stderr)
        )
      raise Exception("Return code not 0, maybe an error! The command was: {}.\n The output of the process was: {}\n The stderr of the process was {}".format(
        command, stdout, stderr))

  def _fast_split(self, *, path, segment_time, path_to_output, prefix_output_name, ext):
    full_path_no_ext = os.path.join(path_to_output, prefix_output_name)
    command = "ffmpeg -i \"{}\" -c copy -map 0:v -segment_time {} -reset_timestamps 1 -f segment {}%03d{}".format(
      os.path.abspath(path), segment_time, full_path_no_ext, ext
    )

    self._call_subprocess(command)

    output_files = list(filter(lambda x: x.startswith(prefix_output_name), os.listdir(path_to_output)))
    return output_files

  def _slow_split(self, *, path, chunk_seconds, path_to_output, prefix_output_name, ext):
    full_path_no_ext = os.path.join(path_to_output, prefix_output_name)
    command = "ffmpeg -i {} -hide_banner  -err_detect ignore_err -segment_time {} -force_key_frames  \"expr: gte(t,n_forced*{})\" -f segment {}%03d{}".format(
      os.path.abspath(path), chunk_seconds, chunk_seconds, full_path_no_ext, ext
    )

    self._call_subprocess(command)

    output_files = list(filter(lambda x: x.startswith(prefix_output_name), os.listdir(path_to_output)))
    return output_files

  def _remove_generated_files(self, files, path_to_output):
    for file in files:
      if os.path.exists(os.path.join(path_to_output, file)):
        os.remove(os.path.join(path_to_output, file))

  def _remove_possible_generated_files(self, prefix_output_name, path_to_output):
    files = list(filter(lambda x: x.startswith(prefix_output_name), os.listdir(path_to_output)))
    for file in files:
      if os.path.exists(os.path.join(path_to_output, file)):
        os.remove(os.path.join(path_to_output, file))

  def _find_smallest_pair(self, files):
    smallest_sum = float('inf')
    for i in range(len(files) - 1):
      pair_sum = files[i][1] + files[i + 1][1]
      if pair_sum < smallest_sum:
        smallest_sum = pair_sum
        smallest_first = i
    return smallest_first

  def _concat_smallest_files(self, files, prefix_output_name, path_to_output, ext):
    # search for the consecutive pair with the smallest sum of durations
    idx = self._find_smallest_pair(files)

    # extract the two smallest files
    f1 = files.pop(idx)
    f2 = files.pop(idx)

    # compute the merged file name
    out_name = prefix_output_name + f1[0].split(prefix_output_name)[1].split(ext)[0] + \
        f2[0].split(prefix_output_name)[1].split(ext)[0]
    out_name += ext

    # compute the full paths of the two files
    f1_full_path = os.path.join(path_to_output, f1[0])
    f2_full_path = os.path.join(path_to_output, f2[0])

    # concat the two smallest files
    out_path = os.path.join(path_to_output, out_name)
    self.concatenate_multiple_video_files([f1_full_path, f2_full_path], out_path)

    # remove the smaller files
    self._remove_generated_files([f1[0], f2[0]], path_to_output)

    # add the resulted file
    files.insert(idx, (out_name, f1[1] + f2[1]))

  def split_video_file(self, path, nr_chunks, path_to_output, duration=None, prefix_output_name=None, default_ext=None):
    # if nr_chunks <= 1:
      # return {'segment_time' : [self.video_file_duration(path=path, as_float=True)], 'output_files' : [path], 'errors': "Splitting 1 movie in {} chunks".format(nr_chunks)}
    # TODO: check if output_file_name is greater than X
    # TODO: check 1k shards
    if prefix_output_name is None:
      prefix_output_name = 'output_part_'

    path_to_output = os.path.abspath(path_to_output)

    _, ext = os.path.splitext(path)
    if default_ext is not None:
      ext = default_ext

    assert len(ext) > 0

    if duration is None or type(duration) == str:
      duration = self.video_file_duration(path=path, as_float=True)
    #endif
    
    chunk_seconds = math.ceil(duration / nr_chunks) + 1
    segment_time = str(timedelta(seconds=chunk_seconds))

    if False:
      try:
        # the faster method
        output_files = self._fast_split(
          path=path,
          segment_time=segment_time,
          path_to_output=path_to_output,
          prefix_output_name=prefix_output_name,
          ext=ext
        )

        durations = [self.video_file_duration(path=os.path.join(
          path_to_output, shard_path), as_float=True) for shard_path in output_files]
        files = list(zip(output_files, durations))

        # split in exactly nr_chunks
        while len(files) > nr_chunks:
          self._concat_smallest_files(files, prefix_output_name, path_to_output, ext)

        output_files = [f[0] for f in files]
        self.P('Finished splitting the files blazingly fast.')

      except Exception as e:
        # the fast approach failed, resorting to the slow approach
        self.P('The fast splitting approach failed on movie {} with exception {}.'.format(path, e))

        self._remove_possible_generated_files(prefix_output_name, path_to_output)
    output_files = self._slow_split(
      path=path,
      chunk_seconds=chunk_seconds,
      path_to_output=path_to_output,
      prefix_output_name=prefix_output_name,
      ext=ext
    )

    output_files.sort()
    dct_ret = {'segment_time': segment_time, 'output_files': output_files, 'errors': _}
    return dct_ret

  def concatenate_multiple_video_files(self, input_paths, output_path):
    output_path = os.path.abspath(output_path)

    if len(input_paths) == 1:
      from shutil import copyfile
      copyfile(input_paths[0], output_path)
      return

    lst_ffmpeg_input_file_rows = ["file '{}'\n".format(os.path.abspath(f)) for f in input_paths]

    tmp = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
    tmp.writelines(lst_ffmpeg_input_file_rows)
    tmp.seek(0)

    command = "ffmpeg -f concat -safe 0 -i {} -c copy {}".format(tmp.name, output_path)

    self._call_subprocess(command)

    tmp.close()
    os.remove(tmp.name)
    return

  def no_frames(self, path):
    command = "ffprobe -v error -count_frames -select_streams v:0 -show_entries stream=nb_read_frames -of default=nokey=1:noprint_wrappers=1 \"{}\"".format(
      path)

    stdout, _ = self._call_subprocess(command)

    return int(stdout)


if __name__ == "__main__":
  video_path = "/home/gts/exe_eng/_local_cache/_data/downloads/20221213170049405891"
  ff = FFMPEGUtils()
  _video_file_duration = ff.video_file_duration(path=video_path)

  dct_ret = ff.split_video_file(
      path=video_path,
      nr_chunks=4,
      path_to_output='/home/gts/exe_eng/_local_cache/_data/downloads/test',
      duration=_video_file_duration,
      default_ext='.mp4'
    )
  print(dct_ret)



import os
import cv2
import numpy as np
import traceback
from naeural_core import constants as ct

from time import sleep
from naeural_core.data.base import DataCaptureThread
from naeural_core.data.mixins_libs import _VideoFileMixin, _VideoConfigMixin

_CONFIG = {
  **DataCaptureThread.CONFIG,
    
  "CAP_RESOLUTION" : ct.CAP_RESOLUTION_NON_LIVE_DEFAULT,
  
  "DISPLAY_STEP"   : 500, # display each 500 frames
  'MAX_RETRIES'           : 2,
  
  "USE_FFMPEG"     : False,

  
  
  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

MAX_FAILED_FRAMES = 10

class VideoFileDataCapture(DataCaptureThread, _VideoFileMixin, _VideoConfigMixin):
  CONFIG = _CONFIG
  def __init__(self, **kwargs):
    self._crt_frame = 0
    self._max_no_frame = MAX_FAILED_FRAMES
    self._failed_frames = 0
    self._resize_h, self._resize_w = None, None
    self._h, self._w = None, None

    super(VideoFileDataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    self.last_url = None

    self._metadata.update(
      fps=None,
      frame_h=None,
      frame_w=None,
      frame_count=None,
      frame_current=None,
      download=None,
      total_time=None,
      time_current=None,
      frame_msec=None,
      path=None,
    )
    return
  
  @property 
  def cfg_c_array(self):
    return self.cfg_stream_config_metadata.get('C_ARRAY', True)

  @property
  def cfg_custom_rotation(self):
    return self.cfg_stream_config_metadata.get('CUSTOM_ROTATION', True)

  @property
  def cfg_delete_path(self):
    delete_path = False
    if not os.path.exists(str(self.video_file_url)):
      delete_path = True

    delete_path = self.cfg_stream_config_metadata.get('DELETE_PATH', delete_path)
    return delete_path

  @property
  def video_file_url(self):
    return self.cfg_url

  def _release(self):
    self._release_cv2_capture()
    
    if self.cfg_delete_path:
      self.P("In `_release` removed file '{}'".format(self._path), color='y')
      os.remove(self._path)
    return

  def _init(self):
    self._force_video_stream_config()
    self._maybe_download()
    return

  def _on_config_changed(self):
    if self.cfg_url != self.last_url:
      self.has_connection = False # forces reconnection
      # TODO(AID): maybe change this log to protect the URL of the video?
      self._maybe_download()
      self.P("URL change detected from {} to {}, reconnecting...".format(
        self.last_url, self.cfg_url,
        )
      )
    return

  def _maybe_reconnect(self):
    if self.has_connection:
      return

    self.has_connection = False
    nr_retry = 1
    str_e = None
    self.last_url = self.cfg_url # a cache of current url
    self._crt_frame = 0
    while nr_retry <= self.cfg_max_retries:
      try:
        self.P('Try #{} connecting to: {} (deq size: {})'.format(nr_retry, self._path, len(self._deque)))

        if not os.path.exists(self._path):
          raise Exception('Connection error. Provided url cannot be found: {}'.format(self._path))

        self._create_cv2_capture()

        if self._cv2_cap is None or not self._cv2_cap.isOpened():
          raise Exception('Video Capture could not be initiated. Please check the url: {}'.format(self._path))

        fps = 20
        try:
          fps = int(self._cv2_cap.get(cv2.CAP_PROP_FPS))
        except OverflowError:
          pass

        
        # height = int(self._cv2_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        # width = int(self._cv2_cap.get(cv2.CAP_PROP_FRAME_WIDTH))


        frame_count = self._get_number_of_frames()

        self._reset_cv2_capture()

        frame_msec = 1000 / fps

        # this makes certain kinds of movie to stop returning frames!!!!
        # self._cv2_cap.set(cv2.CAP_PROP_POS_AVI_RATIO, 1)
        # video_time = self._cv2_cap.get(cv2.CAP_PROP_POS_MSEC)
        # self._cv2_cap.set(cv2.CAP_PROP_POS_AVI_RATIO, 0)

        self.P('FPS: {}, Shape (HxW): {}, Frame count: {}'.format(fps, (self._h, self._w), frame_count), color='g')
        self.has_connection = True

        self._metadata.fps = fps
        self._metadata.finished = False
        self._metadata.frame_h = self._h
        self._metadata.frame_w = self._w
        self._metadata.frame_count = frame_count
        self._metadata.frame_msec = frame_msec
        # self._metadata.total_time = video_time

        ### universal video stream code
        self._maybe_configure_frame_crop_resize(
          height=self._h,
          width=self._w,
        )
        ### end universal video stream code

        if self._metadata.download is None:
          self._metadata.download = 0
        
        msg = 'Connected to stream'
        self.P(msg)
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_NORMAL,
          msg=msg,
          info="{}".format(self._metadata.__dict__),
          displayed=True,
        )
      except:
        str_e = traceback.format_exc()
        self.P('Connect exception for current config {}: \n{}'.format(self.config_data, str_e), color='r')
      # end try-except

      if self.has_connection:
        break
      elif nr_retry == self.cfg_max_retries:
        self.P('Reached maximum number of connect retries. Ending init method.', color='y')
      else:
        sleep(1)

      nr_retry += 1
    # endwhile
    if self.has_connection:
      self.P('Done init')
    else:
      msg = 'Data capture could not be initialized after {} retries'.format(self.cfg_max_retries)
      self.P(msg)
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        info=str_e
      )
    return

  def correct_rotation(self, frame, rotateCode):
    return cv2.rotate(frame, rotateCode)

  def _run_data_aquisition_step(self):
    if self._cv2_cap is not None and self._cv2_cap.isOpened():
      self.start_timer('cv2_read')
      has_frame, frame = self._cv2_cap.read()
      self.end_timer('cv2_read')
      if self.cv2_supports_rotation and  self._cv2_rotate_code is not None and self.cfg_custom_rotation:
        self.start_timer('cv2_rotate')
        frame = self.correct_rotation(frame, self._cv2_rotate_code)
        self.end_timer('cv2_rotate')
      if has_frame:
        ### universal video stream code        
        frame = self._maybe_resize_crop(frame)
        ### end universal video stream code
        # data
        frame = frame[:, :, ::-1]

        if self.cfg_c_array:
          self.start_timer('ascontiguousarray')
          frame = np.ascontiguousarray(frame)
          self.end_timer('ascontiguousarray')
        # endif

        # metadata
        self._crt_frame += 1
        
        if (self._crt_frame % self.cfg_display_step) == 0:
          self.P("Getting frame {}/{} ({:.1f}%)".format(
            self._crt_frame, self._metadata.frame_count,
            self._crt_frame / self._metadata.frame_count * 100,
            ),
          color='g'
          )

        self._metadata.frame_current = self._crt_frame
        self._metadata.time_current = self._cv2_cap.get(cv2.CAP_PROP_POS_MSEC)
        
        if self._crt_frame >= self._metadata.frame_count and not self._metadata.finished:
          self._metadata.finished = True
          self.P("Video stream state is now finished")
        
        self.start_timer('add_video_frame')
        self._add_inputs(
          [
            self._new_input(img=frame, metadata=self._metadata.__dict__.copy()),
          ]
        )
        self.end_timer('add_video_frame')

      else:
        self._failed_frames += 1
        if self._max_no_frame > 0:
          self._max_no_frame -= 1
        else:
          self.has_finished_acquisition = True
          self._crt_frame = 0
          if self._metadata.finished:
            self.P("Empty reads {}/{} => `has_finished_acquisition=True` at {}/{}".format(
              self._failed_frames, MAX_FAILED_FRAMES,
              self._metadata.frame_current,self._metadata.frame_count, 
              ), 
            )
          else:
            self.P("Multiple failed reads at {}/{} => `has_finished_acquisition=True`, Fail/Max: {}/{}".format(
              self._metadata.frame_current,self._metadata.frame_count, 
              self._failed_frames, MAX_FAILED_FRAMES), color='r'
            )
    else:
      self.P('Setting has_connection=False')
      self.has_connection = False
    return


import os
import sys
import json
import pickle
from io import BytesIO, TextIOWrapper

class _PublicTFKerasMixin(object):
  """
  Mixin for public tf keras functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_PickleSerializationMixin`:
    - self.load_pickle_from_models

  * Obs: This mixin uses also attributes/methods of `_JSONSerializationMixin`"
    - self.save_json
    - self.load_json
  """

  def __init__(self):
    super(_PublicTFKerasMixin, self).__init__()

    self.gpu_mem = None
    self.TF = False
    self.TF_VER = None
    self.TF_KERAS_VER = None
    self.KERAS = False
    self.KERAS_VER = None
    self.devices = {}
    return


  def _check_tf_avail(self):
    import tensorflow as tf
    try:
      self.TF_VER = tf.__version__
      found = True
      self.TF = True
    except:
      found = False
      self.TF = False
    return found

  def check_tf(self):
    import tensorflow as tf
    ret = 0
    if self._check_tf_avail():

      CHECK_GPUS = False
      tf_gpu = False
      is_eager = tf.executing_eagerly()
      # tf_gpu = tf.test.is_gpu_available(cuda_only=True)
      self._logger("Found TF {}. [eager mode {}]".format(
        self.TF_VER,
        "ON" if is_eager else "OFF"))
      ret = 2 if tf_gpu else 1
      if CHECK_GPUS:
        from tensorflow.python.client import device_lib
        local_device_protos = device_lib.list_local_devices()
        self.devices = {x.name: {'name': x.physical_device_desc,
                                 'mem': x.memory_limit, }
                        for x in local_device_protos}
        types = [x.device_type for x in local_device_protos]
        self.gpu_mem = []
        if 'GPU' in types:
          ret = 2
          for _dev in self.devices:
            if 'GPU' in _dev.upper():
              self._logger(" {}:".format(_dev[-5:]))
              self._logger("  Name: {}".format(self.devices[_dev]["name"]))
              self.gpu_mem.append(self.devices[_dev]["mem"] / (1024 ** 3))
              self._logger("  Mem:  {:.1f} GB".format(self.gpu_mem[-1]))
      try:
        self.TF_KERAS_VER = tf.keras.__version__
        self._logger("Found TF.Keras {}".format(self.TF_KERAS_VER))
      except:
        self.TF_KERAS_VER = None
        self._logger("No TF.Keras found.")

      if self._check_keras_avail():
        self._logger("Found Keras {}".format(self.KERAS_VER))
    else:
      self._logger("TF not found")
      self.TF = False
    return ret

  def get_gpu(self):
    res = []
    if self._check_tf_avail():
      self.TF = True
      from tensorflow.python.client import device_lib
      loc = device_lib.list_local_devices()
      res = [x.physical_device_desc for x in loc if x.device_type == 'GPU']
    return res

  def _check_keras_avail(self):
    try:
      found = True
      import keras
      self.KERAS_VER = keras.__version__
      self.KERAS = True
    except ImportError:
      found = False
      self.KERAS = False
    return found

  def is_tf2(self):
    if self._check_tf_avail():
      return self.TF_VER[0] == '2'
    else:
      return None

  def _block_tf2(self):
    import tensorflow as tf
    if self.is_tf2():
      if tf.executing_eagerly():
        self.P("WARNING: Called a function/method that uses TF1x approach (sessions, etc) in eager mode")
      else:
        self.P("WARNING: you are using tf2 with tf1 graphs - please ensure you use `tensorflow.compat.v1` - either `import tensorflow.compat.v1 as tf` or in calls")
    return

  def supress_tf_warnings(self):
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    return

  def supress_tf_warn(self):
    return self.supress_tf_warnings()

  def model_exists(self, model_file):
    """
    returns true if model_file (check both .pb or .h5) exists
    """
    exists = False
    for ext in ['', '.h5', '.pb']:
      fpath = os.path.join(self.get_models_folder(), model_file + ext)
      if os.path.isfile(fpath):
        exists = True
    if exists:
      self.P("Detected model {}.".format(fpath))
    else:
      self.P("Model {} NOT found.".format(model_file))
    return exists

  @staticmethod
  def get_keras_model_summary(model, full_info=False):
    if not full_info:
      # setup the environment
      old_stdout = sys.stdout
      sys.stdout = TextIOWrapper(BytesIO(), sys.stdout.encoding)
      # write to stdout or stdout.buffer
      model.summary()
      # get output
      sys.stdout.seek(0)  # jump to the start
      out = sys.stdout.read()  # read output
      # restore stdout
      sys.stdout.close()
      sys.stdout = old_stdout
    else:
      out = model.to_yaml()

    str_result = "Keras Neural Network Layout\n" + out
    return str_result

  def load_keras_model_def_and_weights(self, label, custom_objects=None, subfolder_path=None):
    import tensorflow as tf

    try:
      from custom_layers import CUSTOM_LAYERS
    except:
      try:
        from naeural_core.local_libraries.custom_layers import CUSTOM_LAYERS
      except:
        CUSTOM_LAYERS = {}

    load_folder = self.get_models_folder()
    if subfolder_path is not None:
      load_folder = os.path.join(load_folder, subfolder_path.lstrip('/'))

    if not os.path.exists(load_folder):
      self.P("Source load folder not found: '{}'".format(load_folder))
      return

    path_model_json = os.path.join(load_folder, label + '_model_def.json')
    path_model_weights = os.path.join(load_folder, label + '_weights.h5')

    if not os.path.exists(path_model_json):
      self.P("Model json file for '{}' does not exist in folder '...{}'".format(path_model_json, load_folder[-40:]))
      return

    if not os.path.exists(path_model_weights):
      self.P("Model weights file for '{}' does not exist in folder '...{}'".format(path_model_json, load_folder[-40:]))
      return

    self.P("Loading and reconstructing the model given label '{}' in folder '...{}'".format(
      label, load_folder[-40:]
    ))

    if custom_objects is None:
      custom_objects = {}
    for k, v in CUSTOM_LAYERS.items():
      custom_objects[k] = v

    model_json = json.dumps(self.load_json(path_model_json))

    try:
      model = tf.keras.models.model_from_json(model_json, custom_objects=custom_objects)
    except Exception:
      self.P("ERROR! There was a problem with `tf.keras.models.model_from_json`")
      return

    try:
      model.load_weights(path_model_weights)
    except Exception:
      self.P("ERROR! There was a problem with loading weights")
      return

    self.P("  Successfully loaded and reconstructed the model '{}'".format(label))
    return model

  def load_keras_model_weights(self, filename, model, layers):
    """
    load select 'layers' weights
    """
    assert len(layers) > 0, 'Unknown list of selected layers'
    file_name = os.path.join(self.get_models_folder(), filename + ".pkl")
    if not os.path.isfile(file_name):
      self.P("No weights file found")
      return False
    self.P("Loading weights for {} layers from {}...".format(len(layers), file_name))
    with open(file_name, 'rb') as f:
      w_dict = pickle.load(f)
    self.P("Loaded layers: {}".format(list(w_dict.keys())))
    for layer in layers:
      model.get_layer(layer).set_weights(w_dict[layer])
    self.P("Done loading weights.", show_time=True)
    return True

  def load_keras_model_from_weights(self, weights_file, config_file, custom_objects=None):
    import tensorflow as tf
    from_config = tf.keras.models.Model.from_config
    try:
      from custom_layers import CUSTOM_LAYERS
    except:
      from naeural_core.local_libraries.custom_layers import CUSTOM_LAYERS

    if custom_objects is None:
      custom_objects = {}
    for k, v in CUSTOM_LAYERS.items():
      custom_objects[k] = v

    self.P("Creating model from config and loading weights...")
    if weights_file[-3:] != '.h5':
      weights_file += '.h5'
    if config_file[-4:] != '.pkl':
      config_file += '.pkl'
    fn_cfg = os.path.join(self.get_models_folder(), config_file)
    fn_mdl = os.path.join(self.get_models_folder(), weights_file)
    if not os.path.isfile(fn_cfg):
      raise ValueError("Model config file not found: {}".format(fn_cfg))
    if not os.path.isfile(fn_mdl):
      raise ValueError("Model config file not found: {}".format(fn_mdl))
    dct_cfg = self.load_pickle_from_models(fn_cfg)
    model = from_config(dct_cfg, custom_objects=custom_objects)
    model.load_weights(fn_mdl)
    self.P("  Model created and weights loaded.")
    return model

  # custom_objects: dict with 'custom_name': custom_function
  def load_keras_model(self, model_name, custom_objects=None, DEBUG=True,
                       force_compile=True,
                       full_path=False,
                       subfolder_path=None):
    """
    Wrapper of `tf.keras.models.load_model`

    Parameters
    ----------

    model_name : str
      The name of the model that should be loaded.

    custom_objects : dict, optional
      Custom objects that should be loaded (besides the standard ones -
      from CUSTOM_LAYERS).
      The default is None.

    DEBUG : boolean, optional
      Specifies whether the logging is enabled
      The default is True.

    force_compile : boolean, optional
      `compile` param passed to tf.keras.models.load_model.
      The default is True.

    full_path : boolean, optional
      Specifies whether `model_name` is a full path or should be loaded
      from `_models` folder.
      The default is False.

    subfolder_path : str, optional
      A path relative to '_models' from where the model is loaded
      Default is None.
    """

    try:
      from custom_layers import CUSTOM_LAYERS
    except:
      try:
        from naeural_core.local_libraries.custom_layers import CUSTOM_LAYERS
      except:
        CUSTOM_LAYERS = {}

    if custom_objects is None:
      custom_objects = {}
    for k, v in CUSTOM_LAYERS.items():
      custom_objects[k] = v

    if model_name[-3:] != '.h5':
      model_name += '.h5'
    if DEBUG: self.verbose_log("  Trying to load {}...".format(model_name))

    if not full_path:
      model_folder = self.get_models_folder()
      if subfolder_path is not None:
        model_folder = os.path.join(model_folder, subfolder_path.lstrip('/'))

      model_full_path = os.path.join(model_folder, model_name)
    else:
      model_full_path = model_name

    if os.path.isfile(model_full_path):
      from tensorflow.keras.models import load_model

      if DEBUG: self.verbose_log("  Loading [...{}]".format(model_full_path[-40:]))
      model = load_model(model_full_path, custom_objects=custom_objects, compile=force_compile)
      if DEBUG: self.verbose_log("  Done loading [...{}]".format(model_full_path[-40:]), show_time=True)
    else:
      self.verbose_log("  File {} not found.".format(model_full_path))
      model = None
    return model

  def log_keras_model(self, model):
    self.verbose_log(self.get_keras_model_summary(model))
    return

  def show_model_summary(self, model):
    self.log_keras_model(model)
    return

  def load_graph_from_models(self, model_name, get_input_output=False):
    from naeural_core.local_libraries.nn.tf.utils import load_graph_from_models as _load
    return _load(
      log=self,
      model_name=model_name,
      get_input_output=get_input_output)

  @staticmethod
  def get_tensors_in_tf_graph(tf_graph):
    import tensorflow
    if isinstance(tf_graph, tensorflow.core.framework.graph_pb2.GraphDef):
      return [n.name for n in tf_graph.node]
    else:
      return [n.name for n in tf_graph.as_graph_def().node]



import sys
import platform
import os
import signal
import subprocess
import json
import requests
import psutil

import flask

from functools import partial
from time import sleep, time
from datetime import timedelta

from naeural_core import Logger
from ratio1 import BaseDecentrAIObject
from ratio1.logging.logger_mixins.json_serialization_mixin import NPJson
from naeural_core.local_libraries.model_server_v2.request_utils import get_api_request_body, MSCT

from naeural_core.main.ver import __VER__ as __CORE_VER__

DEFAULT_NR_WORKERS = 5
DEFAULT_HOST = '127.0.0.1'

DEFAULT_SERVER_PATHS = [
  MSCT.RULE_RUN,
  MSCT.RULE_NOTIF,
  MSCT.RULE_UPDATE_WORKERS
]


def get_packages():
  import pkg_resources
  packs = [x for x in pkg_resources.working_set]
  maxlen = max([len(x.key) for x in packs]) + 1
  packs = [
    "{}{}".format(x.key + ' ' * (maxlen - len(x.key)), x.version) for x in packs
  ]
  packs = sorted(packs)  
  return packs  

class FlaskGateway(BaseDecentrAIObject):

  app = None

  def __init__(self, log : Logger,
               workers_location,
               server_names=None,
               workers_suffix=None,
               host=None,
               port=None,
               first_server_port=None,
               server_execution_path=None,
               **kwargs
              ):

    """
    Parameters:
    -----------
    log : Logger, mandatory

    workers_location: str, mandatory
      Dotted path of the folder where the business logic of the workers is implemented

    server_names: List[str], optional
      The names of the servers that will be run when the gateway is opened. This names should be names of .py files
      found in `workers_location`
      The default is None (it takes all the keys in MSCT.CONFIG_ENDPOINTS)

    workers_suffix: str, optional
      For each worker, which is the suffix of the class name.
      e.g. if the worker .py file is called 'get_similarity.py', then the name of the class is GetSimilarity<Suffix>.
      If `workers_suffix=Worker`; then the name of the class is GetSimilarityWorker
      The default is None

    host: str, optional
      Host of the gateway
      The default is None ('127.0.0.1')

    port: int, optional
      Port of the gateway
      The default is None (5000)

    first_server_port: int, optional
      Port of the first server (the ports are allocated sequentially starting with `first_port_server`)
      The default is None (port+1)

    server_execution_path: str, optional
      The API rule where the worker logic is executed.
      The default is None (MSCT.RULE_DEFAULT)
    """

    self.__version__ = __CORE_VER__

    self._start_server_names = server_names
    self._host = host or '127.0.0.1'
    self._port = port or 5000
    self._server_execution_path = server_execution_path or MSCT.RULE_DEFAULT
    self._workers_location = workers_location
    self._workers_suffix = workers_suffix

    self._first_server_port = first_server_port or self._port + 1
    self._current_server_port = self._first_server_port

    self._config_endpoints = None
    
    self._start_time = time()

    self._servers = {}
    self._paths = None
    super(FlaskGateway, self).__init__(log=log, prefix_log='[FSKGW]', **kwargs)
    return
  
  
  def get_response(self, data):
    if MSCT.DOWNLOAD_FILE_COMMAND in data:
      # MSCT.DOWNLOAD_FILE_COMMAND is a special field in responses that allows the creation of a file download from the
      # gateway directly to the client
      self.P("Received {} command:\n{}".format(MSCT.DOWNLOAD_FILE_COMMAND, json.dumps(data, indent=4)))
      dct_download = data[MSCT.DOWNLOAD_FILE_COMMAND]
      fn = dct_download[MSCT.DOWNLOAD_FILE_PATH]
      return flask.send_file(fn)
    else:
      if isinstance(data, dict):
        dct_result = data
      else:
        dct_result = {
          MSCT.DATA : data,        
        }
      dct_result[MSCT.VER] = __CORE_VER__
      dct_result[MSCT.TIME] = self.log.time_to_str()
      dct_result[MSCT.GW_UPTIME] = self._elapsed_to_str(time() - self._start_time)
      return flask.jsonify(dct_result)

  def _log_banner(self):
    _logo = "FlaskGateway v{} started on '{}:{}'".format(
      self.__version__, self._host, self._port
    )

    lead = 5
    _logo = " " * lead + _logo
    s2 = (len(_logo) + lead) * "*"
    self.log.P("")
    self.log.P(s2)
    self.log.P("")
    self.log.P(_logo)
    self.log.P("")
    self.log.P(s2)
    self.log.P("")
    return

  def startup(self):
    super().startup()
    self._log_banner()
    self._no_startup_wait = self.config_data.get(MSCT.NO_STARTUP_WAIT, False)
    self._config_endpoints = self.config_data.get(MSCT.CONFIG_ENDPOINTS, {})

    if self._start_server_names is None:
      self._start_server_names = list(self._config_endpoints.keys())

    if not self._server_execution_path.startswith('/'):
      self._server_execution_path = '/' + self._server_execution_path

    self.start_servers()

    if self._paths is None:
      self.kill_servers()
      raise ValueError("Gateway cannot start because no paths were retrieved from endpoints.")

    self.app = flask.Flask('FlaskGateway')
    self.app.json_encoder = NPJson
    for rule in self._paths:
      partial_view_func = partial(self._view_func_worker, rule)
      partial_view_func.__name__ = "partial_view_func_{}".format(rule.lstrip('/'))
      self.P("Registering {} on `{}`".format(rule, partial_view_func.__name__), color='g')
      self.app.add_url_rule(
        rule=rule,
        view_func=partial_view_func,
        methods=['GET', 'POST', 'OPTIONS']
      )
    #endfor

    self.app.add_url_rule(
      rule=MSCT.RULE_START,
      endpoint='StartServerEndpoint',
      view_func=self._view_func_start_server,
      methods=['GET', 'POST']
    )

    ### THIS SHOULD BE USED WITH CARE IN PROD
    if True:
      self.app.add_url_rule(
        rule=MSCT.RULE_SHUTDOWN,
        endpoint='ShutdownGateway',
        view_func=self._view_shutdown,
        methods=['GET', 'POST']
      )

    self.app.add_url_rule(
      rule=MSCT.RULE_KILL,
      endpoint='KillServerEndpoint',
      view_func=self._view_func_kill_server,
      methods=['GET', 'POST']
    )
    
    self.app.add_url_rule(
      rule=MSCT.RULE_LIST,
      endpoint='ListServersEndpoint',
      view_func=self._view_list_servers,
      methods=['GET', 'POST']
    )

    self.app.add_url_rule(
      rule=MSCT.RULE_SYS,
      endpoint='SystemHealthStatus',
      view_func=self._view_system_status,
      methods=['GET', 'POST']
    )

    self.P("Starting gateway server after all endpoints have been defined...", color='g')
    self._get_system_status(display=True)
    self.app.run(
      host=self._host,
      port=self._port,
      threaded=True
    )
    return
  
  def _get_system_status(self, display=True):
    mem_total = round(self.log.get_machine_memory(gb=True),2)
    mem_avail = round(self.log.get_avail_memory(gb=True),2)
    mem_gateway = round(self.log.get_current_process_memory(mb=False),2)
    mem_servers = 0
    dct_servers = {
    }
    for svr in self._servers:
      proc = psutil.Process(self._servers[svr][MSCT.PROCESS].pid)
      proc_mem = round(proc.memory_info().rss / (1024**3), 2)
      mem_servers += proc_mem
      dct_servers[svr] = proc_mem
    #endfor calc mem
    mem_used = round(mem_gateway + mem_servers, 2)
    mem_sys = round((mem_total - mem_avail) - mem_used,2)
    self.P("  Total server memory:    {:>5.1f} GB".format(mem_total), color='g')
    self.P("  Total server avail mem: {:>5.1f} GB".format(mem_avail), color='g')
    self.P("  Total allocated mem:    {:>5.1f} GB".format(mem_used), color='g')
    self.P("  System allocated mem:   {:>5.1f} GB".format(mem_sys), color='g')
    dct_stats = dict(
      mem_total=mem_total,
      mem_avail=mem_avail,
      mem_gateway=mem_gateway,
      mem_used=mem_used,
      mem_sys=mem_sys,
      mem_servers=dct_servers,
      system=platform.platform(),
      py=sys.version,
      packs=get_packages(),
      info='Memory Size is in GB. Total and avail mem may be reported inconsistently in containers.'
    )
    return dct_stats


  def _start_server(self, server_name, port, execution_path, host=None, nr_workers=None, verbosity=1):
    config_endpoint = self._config_endpoints.get(server_name, {})
    
    desc = config_endpoint.get(MSCT.DESCRIPTION)
    self.P('Attempting to start server with "SIGNATURE" : "{}"'.format(server_name))
    self.P('  Description: "{}"'.format(desc))
    
    if config_endpoint.get(MSCT.DISABLED):
      self.P("Skipping server '{}' due to its {} status: {}".format(server_name, MSCT.DISABLED, config_endpoint))
      return False
    
    if MSCT.SERVER_CLASS in config_endpoint:
      server_class = config_endpoint[MSCT.SERVER_CLASS]
      self.P("Found {} '{}' in endpoint '{}' definition - using this class as worker".format(
        MSCT.SERVER_CLASS, server_class, server_name,
      ))
    else:
      server_class = server_name

    if MSCT.HOST in config_endpoint:
      host = config_endpoint[MSCT.HOST]
    else:
      host = host or DEFAULT_HOST
      self.P("WARNING: '{}' not provided in endpoint configuration for {}.".format(MSCT.HOST, server_name), color='r')
    #endif

    if MSCT.NR_WORKERS in config_endpoint:
      nr_workers = config_endpoint[MSCT.NR_WORKERS]
    else:
      nr_workers = nr_workers or DEFAULT_NR_WORKERS
      self.P("WARNING: MSCT.NR_WORKERS not provided in endpoint configuration for {}.".format(server_name), color='r')
    #endif

    msg = "Creating server `{} <{}>` at {}:{}{}".format(server_name, server_class, host, port, execution_path)
    self.P(msg, color='g')
    self._create_notification('log', msg)

    popen_args = [
      'python',
      'libraries/model_server_v2/run_server.py',
      '--base_folder', self.log.root_folder,
      '--app_folder', self.log.app_folder,
      '--config_endpoint', json.dumps(config_endpoint),
      '--host', host,
      '--port', str(port),
      '--execution_path', execution_path,
      '--workers_location', self._workers_location,
      '--worker_name', server_class,
      '--worker_suffix', self._workers_suffix,
      '--microservice_name', server_name,
      '--nr_workers', str(nr_workers),
      '--use_tf',
    ]

    process = subprocess.Popen(popen_args)

    self._servers[server_name] = {
      MSCT.PROCESS   : process,
      MSCT.HOST      : host,
      MSCT.PORT      : port,
      MSCT.START     : time(),
    }

    sleep(1)

    msg = "Successfully created server '{}' with PID={}".format(server_name, process.pid)
    self.P(msg, color='g')
    self._create_notification('log', msg)
    #endif

    return True
  
  def _elapsed_to_str(self, t):
    return str(timedelta(seconds=int(t)))
  
  def _get_server_status(self, server_name):
    online = False
    urls = []
    _error = None
    paths = None
    try:
      url = 'http://{}:{}{}'.format(
        self._servers[server_name][MSCT.HOST],
        self._servers[server_name][MSCT.PORT],
        MSCT.RULE_PATHS
      )
      response = requests.get(url=url)
      paths = response.json()[MSCT.PATHS]
      urls = [url]
      for path in paths:
        url = 'http://{}:{}{}'.format(
          self._servers[server_name][MSCT.HOST],
          self._servers[server_name][MSCT.PORT],
          path
        )
        urls.append(url)
      online = True
    except Exception as exc:
      _error = str(exc)
  
    result = {
      MSCT.ONLINE  : online,
      MSCT.ERROR   : _error,
      MSCT.URLS    : urls,
      MSCT.PATHS   : paths,
      MSCT.PORT    : self._servers[server_name][MSCT.PORT],
      MSCT.UPTIME  : self._elapsed_to_str(time() - self._servers[server_name][MSCT.START])
    }
    return result

  def _get_paths_from_server(self, server_name):
    self.P("Requesting `get_paths` from server '{}' in order to map available paths...".format(server_name))
    
    resp = self._get_server_status(server_name)
    if not resp[MSCT.ONLINE]:
      raise ValueError('Server not yet online: {}'.format(resp[MSCT.ERROR]))
    else:
      self._paths = resp[MSCT.PATHS]
      self.P("  Responded with paths={}".format(self._paths), color='g')
    return

  def start_servers(self):
    for i,server_name in enumerate(self._start_server_names):
      if self._start_server(
        server_name=server_name,
        port=self._current_server_port,
        execution_path=self._server_execution_path,
        verbosity=1
      ):
        self._current_server_port += 1
    #endfor

    if self._no_startup_wait:
      self.P("Fast startup enabled, using default paths: {}".format(DEFAULT_SERVER_PATHS), color='g')
      self._paths = DEFAULT_SERVER_PATHS 
    else:
      nr_tries = 0    
      svr = self.config_data.get(MSCT.DEFAULT_SERVER,  self._start_server_names[0])
      self.P("")
      WAIT_ON_ERROR = 10
      while True:
        try:
          nr_tries += 1
          self._get_paths_from_server(svr)
          self.P("  Done getting paths.", color='g')
          break
        except Exception as exc:
          self.P("  Error: {}".format(exc), color='r')
          if nr_tries >= (120 / WAIT_ON_ERROR):
            raise ValueError("Could not get paths from server '{}'".format(svr))
          sleep(WAIT_ON_ERROR)
        #end try-except
      #endwhile
    #endif no startup wait or wait for paths

    return


  def _get_server_process(self, server_name):
    return self._servers[server_name][MSCT.PROCESS]

  def _server_exists(self, server_name):
    return server_name in self._servers

  @property
  def active_servers(self):
    return list(self._servers.keys())

  def _kill_server_by_name(self, server_name):
    TIMEOUT = 5
    process = self._get_server_process(server_name)
    process.terminate()
    process.wait(TIMEOUT)
    if process.returncode is None:
      self.P("Terminating '{}:{}' with kill signal after {}s".format(
        server_name, process.pid, TIMEOUT))
      process.kill()
      sleep(1)
    self.P("  '{}' terminated with code: {}".format(server_name, process.returncode))
    self._servers.pop(server_name)
    return

  def kill_servers(self):
    names = list(self._servers.keys())
    for server_name in names:
      self.P("Terminating server '{}' ...".format(server_name))
      self._kill_server_by_name(server_name)
      sleep(2)
      self.P("  Server '{}' deallocated.".format(server_name))
    return
  

  def _view_func_worker(self, path):
    request = flask.request
    params = get_api_request_body(request, self.log)
    signature = params.pop(MSCT.SIGNATURE, None)
    if signature is None:
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : "Bad input. MSCT.SIGNATURE not found"})

    if signature not in self._servers:
      return self.get_response({
        MSCT.VER : __CORE_VER__, 
        MSCT.ERROR : "Bad signature {}. Available signatures/servers: {}".format(
          signature, 
          self.active_servers
        )
      })

    url = 'http://{}:{}{}'.format(
      self._servers[signature][MSCT.HOST],
      self._servers[signature][MSCT.PORT],
      path
    )

    response = requests.post(url, json=params)
    return self.get_response(response.json())

  def _view_func_start_server(self):
    request = flask.request
    params = get_api_request_body(request, self.log)
    signature = params.get(MSCT.SIGNATURE, None)

    if signature is None:
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : f"Bad input. {MSCT.SIGNATURE} not found"})

    if self._server_exists(signature):
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : "Signature {} already started".format(signature)})

    resp = self._start_server(
      server_name=signature,
      port=self._current_server_port,
      execution_path=self._server_execution_path,
      verbosity=0
    )
    if resp:
      self._current_server_port += 1
      return self.get_response({'MESSAGE': 'OK.'})
    else:
      return self.get_response({'MESSAGE': 'Server DISABLED.'})

  def _view_func_kill_server(self):
    request = flask.request
    params = get_api_request_body(request, self.log)
    signature = params.get(MSCT.SIGNATURE, None)

    if signature is None:
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : f"Bad input. {MSCT.SIGNATURE} not found"})
    
    if signature == '*':
      self.kill_servers()      
    elif not self._server_exists(signature):
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : "Bad signature {}. Available signatures: {}".format(signature, self.active_servers)})
    else:
      process = self._get_server_process(signature)
      self._kill_server_by_name(signature)
      return self.get_response({'MESSAGE' : 'OK. Killed PID={} with return_code {}.'.format(
        process.pid,
        process.returncode
      )})


  def _view_list_servers(self):
    return self.get_response({
      MSCT.AVAIL_SERVERS : {
        svr_name : self._get_server_status(svr_name)
        for svr_name in self._servers
      },
      MSCT.VER : __CORE_VER__,
    })
  
  
  def _view_system_status(self):
    return self.get_response({
      MSCT.SYSTEM_STATUS : self._get_system_status(display=True)
    })
    
    


  def _view_shutdown(self):
    request = flask.request
    params = get_api_request_body(request, self.log)
    signature = params.get(MSCT.SIGNATURE, None)

    if signature is None:
      return self.get_response({MSCT.VER : __CORE_VER__, MSCT.ERROR : f"Bad input. {MSCT.SIGNATURE} not found"})
    
    if signature.upper() == MSCT.KILL_CMD:
      self.kill_servers()
      _pid = os.getpid()
      _signal = signal.SIGKILL
      self.P("Terminating gateway server v{} with pid {} with signal {}...".format(__VER__, _pid, _signal))
      os.kill(_pid, _signal)
      self.P("Running _exit() ...")
      os._exit(1)

    if not self._server_exists(signature):
      return self.get_response({MSCT.ERROR : "Bad signature {}. Available signatures: {}".format(signature, self.active_servers)})




import pandas as pd
from naeural_core import constants as ct

from naeural_core.data.base import DataCaptureThread

_CONFIG = {
  **DataCaptureThread.CONFIG,
  'VALIDATION_RULES' : {
    **DataCaptureThread.CONFIG['VALIDATION_RULES'],
  },
}

class StructuredSequenceFileDataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):

    self._initial_data = None

    super(StructuredSequenceFileDataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    # NON-threaded code in startup
    super().startup()    

    self.P("Preparing INIT training data...")
    url_initial_data = self.config.get(ct.INIT_DATA, None)
    if url_initial_data is None:
      self.P("{} URL is missing!".format(ct.INIT_DATA), color='r')
      raise ValueError("{} failed at startup".format(self.__class__.__name__))

    fn = self._download(
      url=url_initial_data,
      progress=True
    )

    df = pd.read_csv(fn)
    self._initial_data = df.to_dict()
    self.P("Done preparing INIT data.")

    self._metadata.update(
      train_url=url_initial_data,
      train_data_shape=df.shape,
      test_data_url=self.cfg_url,
      dataframe_current=0,
      dataframe_count=0,
      current_pass=0,
    )
    return  
  
  def _init(self):
    ### TODO why don't we move the training data download here?!

    self.P('Downloading test-time data...')
    try:
      # following code will be replaced with connnection to IoT device 
      # for non-file data plugins
      path = self._download(
        url=self.cfg_url,
        progress=False,
      )
      self.P('  File downloaded in: {}'.format(path))
      self._df = pd.read_csv(path)
      self.frame_current = 0
      self._metadata.dataframe_count = self._df.shape[0]
      self._current_pass = 0
      self.has_connection = True
      self._generate_data_payload(data=None, initial_data=self._initial_data)
    except:
      msg = "Exception download"
      self.P(msg, color='r')
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
        msg=msg,
        autocomplete_info=True
      )
    return
  
  def _maybe_reconnect(self):
    if self.has_connection:
      return    
    self.frame_current = 0
    self.has_connection = True
    self._current_pass += 1
    return
  
  def _generate_data_payload(self, data=None, initial_data=None):
    self._add_inputs(
      [
        self._new_input(img=None, struct_data=data, metadata=self._metadata.__dict__.copy(), init_data=initial_data)
      ]
    )

    return
  
  def _run_data_aquisition_step(self):
    if self.frame_current >= self._metadata.dataframe_count:
      self.P("Abnormal read reached!")
      self.has_finished_acquisition = True
      return
    dct_rec = self._df.iloc[self.frame_current].to_dict()
    self._metadata.dataframe_current = self.frame_current
    self._metadata.current_pass = self._current_pass
    self._generate_data_payload(data=dct_rec, initial_data=None)
    
    self.frame_current += 1
    if self.frame_current > self._metadata.dataframe_count:
      self.has_finished_acquisition = True
    return 
    
  def _release(self):
    return
  


import importlib
import os
import shutil
import tempfile

from jinja2 import Environment, FileSystemLoader

from naeural_core.business.base.web_app.base_web_app_plugin import BaseWebAppPlugin as BasePlugin
from naeural_core.utils.uvicorn_fast_api_ipc_manager import get_server_manager
from naeural_core.utils.fastapi_utils import PostponedRequest

#TODO: move __sign and __get_response from dauth_manager to base_web_app_plugin or fastapi_web_app or utils
#  all responses should contain the data from __get_response

__VER__ = '0.0.0'

_CONFIG = {
  **BasePlugin.CONFIG,
  'NGROK_ENABLED': True,
  'NGROK_DOMAIN': None,
  'NGROK_EDGE_LABEL': None,

  'PORT': None,

  'ASSETS': None,
  'JINJA_ARGS': {},
  'TEMPLATE': 'basic_server',

  'API_TITLE': None,  # default is plugin signature
  'API_SUMMARY': None,  # default is f"FastAPI created by {plugin signature}"
  'API_DESCRIPTION': None,  # default is plugin docstring

  'PAGES': [],
  'STATIC_DIRECTORY': 'assets',
  'DEFAULT_ROUTE': None,

  # In case of wrapped response, the response will be wrapped in a json with 2 keys:
  # 'result' and 'node_addr', where 'result' is the actual response and 'node_addr' is the node address
  # In case of raw response, the response will be the actual response provided by the endpoint method.
  # The default is 'WRAPPED'.
  'RESPONSE_FORMAT': 'WRAPPED',

  "LOG_REQUESTS": False,

  'PROCESS_DELAY': 0,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES']
  },
}


class FastApiWebAppPlugin(BasePlugin):
  """
  A plugin which exposes all of its methods marked with @endpoint through
  fastapi as http endpoints.

  The @endpoint methods can be triggered via http requests on the web server
  and will be processed as part of the business plugin loop.
  """

  CONFIG = _CONFIG
  
  @property
  def api_title(self):
    return repr(self.cfg_api_title or self.get_signature())
  
  @property
  def uvicorn_server_started(self):
    return self.__uvicorn_server_started

  @staticmethod
  def endpoint(func=None, *, method="get", require_token=False):
    """
    Decorator that marks a method as an HTTP endpoint. Optionally enforces a Bearer token.
    
    Parameters
    ----------
    method : str
        HTTP method (e.g. "get", "post").
        
    require_token : bool
        Whether this endpoint should require a Bearer token.
        
    """
    if func is None:
      def wrapper(func):
        return FastApiWebAppPlugin.endpoint(func, method=method, require_token=require_token)

      return wrapper

    func.__endpoint__ = True
    if isinstance(method, str):
      method = method.lower()
    func.__http_method__ = method
    func.__require_token__ = require_token
    return func

  def get_web_server_path(self):
    return self.script_temp_dir

  def get_package_base_path(self, package_name):
    """
    Return the file path of an installed package parent directory.
    This method was copied from the _PluginsManagerMixin class from ratio1 SDK.

    Parameters
    ----------
    package_name : str
        The name of the installed package.

    Returns
    -------
    str
        The path to the package parent.
    """
    spec = importlib.util.find_spec(package_name)
    if spec is not None and spec.submodule_search_locations:
      return os.path.dirname(spec.submodule_search_locations[0])
    else:
      self.P("Package '{}' not found.".format(package_name), color='r')
    return None

  def validate_static_directory(self):
    """
    Validate the `STATIC_DIRECTORY` value.
    This will check if the value is trying to reference a restricted directory.
    """
    # Creating a temporary directory to check if the static directory is valid
    tmp_directory = tempfile.mkdtemp()
    static_directory = self.cfg_static_directory

    full_path = os.path.join(tmp_directory, static_directory)
    # Check if the full path is a subdirectory of the temporary directory
    if not full_path.startswith(tmp_directory):
      # If the full path is not a subdirectory of the temporary directory, it is invalid
      self.add_error(f"Invalid `STATIC_DIRECTORY`: {static_directory}")
    # endif path is restricted
    # Clean up the temporary directory
    os.rmdir(tmp_directory)
    return

  def validate_default_route(self):
    """
    Validate the `DEFAULT_ROUTE` value.
    It should be a valid URL path if specified.

    """
    default_route = self.cfg_default_route
    if default_route is None:
      # If not provided, return without validation
      return

    if not isinstance(default_route, str):
      self.add_error(f"Invalid `DEFAULT_ROUTE`: {default_route}. It should be a string")
      return

    if not default_route.startswith('/'):
      self.add_error(f"Invalid `DEFAULT_ROUTE`: {default_route}. It should start with '/'")
    # endif default route is invalid
    return

  def initialize_assets(self, src_dir, dst_dir, jinja_args):
    """
    Initialize and copy fastapi assets, expanding any jinja templates.
    All files from the source directory are copied to the
    destination directory with the following exceptions:
      - are symbolic links are ignored
      - files named ending with .jinja are expanded as jinja templates,
        .jinja is removed from the filename and the result copied to
        the destination folder.
    This maintains the directory structure of the source folder.
    In case src_dir is None, only the jinja templates are expanded.

    Parameters
    ----------
    src_dir: str or None, path to the source directory
    dst_dir: str, path to the destination directory
    jinja_args: dict, jinja keys to use while expanding the templates

    Returns
    -------
    None
    """
    self.prepared_env['PYTHONPATH'] = '.:' + os.getcwd() + ':' + self.prepared_env.get('PYTHONPATH', '')

    super(FastApiWebAppPlugin, self).initialize_assets(src_dir, dst_dir, jinja_args)

    package_base_path = self.get_package_base_path('naeural_core')
    if package_base_path is None:
      self.P("Skipping `main.py` rendering, package 'naeural_core' not found.", color='r')
      self.failed = True
      return
    # endif package base path not found
    static_directory = self.jinja_args.get('static_directory')

    if self.cfg_template is not None:
      env = Environment(loader=FileSystemLoader(package_base_path))

      # make sure static directory folder exists
      os.makedirs(self.os_path.join(dst_dir, static_directory), exist_ok=True)

      # Finally render main.py
      template_dir = self.os_path.join('naeural_core', 'business', 'base', 'uvicorn_templates')
      app_template = self.os_path.join(template_dir, f'{self.cfg_template}.j2')
      # env.get_template expects forward slashes, even on Windows.
      app_template = app_template.replace(os.sep, '/')
      app_template = env.get_template(app_template)
      rendered_content = app_template.render(jinja_args)

      with open(self.os_path.join(dst_dir, 'main.py'), 'w') as f:
        f.write(rendered_content)
    # endif render main.py

    # Here additional generic assets can be added if needed
    favicon_path = self.os_path.join(package_base_path, 'naeural_core', 'utils', 'web_app', 'favicon.ico')
    favicon_dst = self.os_path.join(dst_dir, static_directory, 'favicon.ico')
    if self.os_path.exists(favicon_path):
      self.P(f'Copying favicon from {favicon_path} to {favicon_dst}')
      os.makedirs(self.os_path.dirname(favicon_dst), exist_ok=True)
      shutil.copy2(favicon_path, favicon_dst)
    # endif favicon exists

    return

  def _init_endpoints(self) -> None:
    """
    Populate the set of jinja arguments with values needed to create http
    endpoints for all methods of the plugin marked with @endpoint. Since
    there should be at least one such method, this method is always invoked
    via the on_init hook

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    import inspect
    self._endpoints = {}
    jinja_args = []

    def _filter(obj):
      try:
        return inspect.ismethod(obj)
      except Exception as _:
        pass
      return False

    for name, method in inspect.getmembers(self, predicate=_filter):
      if not hasattr(method, '__endpoint__'):
        continue
      self._endpoints[name] = method
      http_method = method.__http_method__
      require_token = getattr(method, '__require_token__', False)
      signature = inspect.signature(method)
      doc = method.__doc__ or ''
      all_params = [param.name for param in signature.parameters.values()]
      all_args = [str(param) for param in signature.parameters.values()]
      if self.cfg_debug_web_app:
        self.P(f'Endpoint {name}[{require_token=}] with {all_args=} and {all_params=}')
      if not require_token:
        args = all_args
        params = all_params
      else:
        if all_params[0] != 'token':
          raise ValueError(f"First parameter of method {name} must be 'token' if require_token is True.")
        params = all_params[1:]
        args = all_args[1:]
        if self.cfg_debug_web_app:
          self.P(f'Endpoint {name}[{require_token=}] left with {args=} and {params=}')
      # endif require_token   
      jinja_args.append({
        'name': name,
        'method': http_method,
        'args': args,
        'params': params,
        'endpoint_doc': doc,
        'require_token': require_token
      })
      str_function = f"{name}({', '.join(args)})"
      self.P(f"Registered endpoint {str_function} with method {http_method}. Require token: {require_token}")
    # endfor all methods
    self._node_comms_jinja_args = jinja_args
    return

  def on_init(self):
    self.__uvicorn_server_started = False
    # Register all endpoint methods.
    self._init_endpoints()
    
  

    # FIXME: move to setup_manager method
    self.manager_auth = b'abc'
    self._manager = get_server_manager(self.manager_auth)
    self.postponed_requests = self.deque()

    self.P("manager address: {}", format(self._manager.address))
    _, self.manager_port = self._manager.address

    # Start the FastAPI app
    self.P('Starting FastAPI app...')
    super(FastApiWebAppPlugin, self).on_init()
    return

  def create_postponed_request(self, solver_method, method_kwargs={}):
    """
    Create a postponed request to be processed by the plugin in the next loop.
    Parameters
    ----------
    solver_method : method
        The method that will solve the postponed request.
    method_kwargs : dict
        The keyword arguments to be passed to the solver_method.
    Returns
    -------
    res : PostponedRequest
        The postponed request object.
    """
    return PostponedRequest(
      solver_method=solver_method,
      method_kwargs=method_kwargs
    )

  def get_postponed_dict(self, request_id, request_value, endpoint_name):
    return {
      'id': request_id,
      'value': request_value,
      'endpoint_name': endpoint_name
    }

  def parse_postponed_dict(self, request):
    return request['id'], request['value'], request['endpoint_name']

  def __fastapi_process_response(self, response):
    if self.cfg_response_format == 'RAW':
      return response
    additional_data = self.get_additional_env_vars()
    additional_data = {
      (k.lower() if isinstance(k, str) else k): v for k, v in additional_data.items()
    }
    return {
      'result': response,
      # TO BE REMOVED maybe (they are also in the additional_data, but maybe the client expects these keys)
      'server_node_addr': self.e2_addr,
      'evm_network' : self.evm_network,
      **additional_data
    }

  def __fastapi_handle_response(self, id, value):
    # TODO: add here message signing
    response = {
      'id': id,
      'value': self.__fastapi_process_response(value)
    }
    self._manager.get_client_queue().put(response)
    return

  def _process(self):
    super(FastApiWebAppPlugin, self)._process()
    new_postponed_requests = []
    while len(self.postponed_requests) > 0:
      request = self.postponed_requests.popleft()
      id, value, endpoint_name = self.parse_postponed_dict(request)

      method = value.get_solver_method()
      kwargs = value.get_method_kwargs()

      try:
        value = method(**kwargs)
      except Exception as exc:
        self.P(
          f'Exception occurred while processing postponed request for {endpoint_name} with method {method.__name__} '
          f'and args:\n{kwargs}\nException:\n{self.get_exception()}',
          color='r'
        )
        value = {
          'error': str(exc)
        }

      if isinstance(value, PostponedRequest):
        new_postponed_requests.append(self.get_postponed_dict(
          request_id=id,
          request_value=value,
          endpoint_name=endpoint_name
        ))
      else:
        if self.cfg_log_requests:
          self.P(f"Request {id} for {method} processed.")
        self.__fastapi_handle_response(id, value)
      # endif request is postponed
    # end while there are postponed requests
    for request in new_postponed_requests:
      self.postponed_requests.append(request)
    # endfor all new postponed requests
    while not self._manager.get_server_queue().empty():
      request = self._manager.get_server_queue().get()
      id = request['id']
      value = request['value']

      method = value[0]
      args = value[1:]

      try:
        if self.cfg_log_requests:
          self.P(f"Received request {id} for {method}.")
        value = self._endpoints.get(method)(*args)
      except Exception as exc:
        self.P("Exception occured while processing\n"
               "Request: {}\nArgs: {}\nException:\n{}".format(
                   method, args, self.get_exception()), color='r')
        value = {
          'error': str(exc)
        }

      if isinstance(value, PostponedRequest):
        self.P(f"Postponing request {id} for {method}.")
        self.postponed_requests.append(self.get_postponed_dict(
          request_id=id,
          request_value=value,
          endpoint_name=method
        ))
      else:
        if self.cfg_log_requests:
          self.P(f"Request {id} for {method} processed.")
        self.__fastapi_handle_response(id, value)
      # endif request is postponed
    # end while

    return None

  def on_close(self):
    self._manager.shutdown()
    super(FastApiWebAppPlugin, self).on_close()
    return

  def __get_uvicorn_process_args(self):
    return f"uvicorn --app-dir {self.script_temp_dir} main:app --host 0.0.0.0 --port {self.port}"

  def get_default_description(self):
    return self.__doc__
  
  
  def on_log_handler(self, text):
    STARTUP_MSG = "Uvicorn running on "
    if STARTUP_MSG in text:
      self.__uvicorn_server_started = True
      match = self.re.search(r"running on (.*?) \(Press", text)
      found = match.group(1) if match else ""
      self.P(f"Server {self.api_title} started on {found}", boxed=True)
    return

  @property
  def jinja_args(self):
    cfg_jinja_args = self.deepcopy(self.cfg_jinja_args)

    dct_pages = cfg_jinja_args.pop('html_files', self.cfg_pages)
    for page in dct_pages:
      page['method'] = 'get'

    static_directory = cfg_jinja_args.pop('static_directory', self.cfg_static_directory)
    default_route = self.cfg_default_route

    return {
      'static_directory': static_directory,
      'html_files': dct_pages,
      'manager_port': self.manager_port,
      'manager_auth': self.manager_auth,
      'api_title': self.api_title,
      'api_summary': repr(self.cfg_api_summary or f"Ratio1 WebApp created with {self.get_signature()} plugin"),
      'api_description': repr(self.cfg_api_description or self.get_default_description()),
      'api_version': repr(self.__version__),
      'node_comm_params': self._node_comms_jinja_args,
      'debug_web_app': self.cfg_debug_web_app,
      'default_route': default_route,
      **cfg_jinja_args,
    }

  def get_start_commands(self):
    super_start_commands = super(FastApiWebAppPlugin, self).get_start_commands()
    return super_start_commands + [self.__get_uvicorn_process_args()]



from time import time, sleep
from naeural_core import constants as ct


class _DefaultCommMixin(object):

  def __init__(self):
    self._last_read = time()
    super(_DefaultCommMixin, self).__init__()
    return
  
  def _post_on_message(self):
    self.P("Received config request: {}...".format(str(self._recv_buff[-1])[:250]))
    return

  def _run_thread_default(self):
    self._init()
    bytes_delivered = 1 # force to 1 to trigger the first send
    while True:
      try:
        # check stop
        start_it = time()
        if self._stop:
          # stop command received from outside. stop imediatly
          self.P('`stop` command received. Exiting from `{}._run_thread_default`'.format(
            self.__class__.__name__))
          break

        # handle send
        self._maybe_reconnect_send()
        if bytes_delivered > 0:
          to_send = None
          if len(self._send_buff) > 0:
            to_send = self._send_buff.popleft()
            # if to_send is None then set bytes_delivered to 1 (1 byte bytes_delivered means error)
            bytes_delivered = 0 if to_send is not None else 1 

        if to_send is not None and self.has_send_conn:
          # msg is always a dict!
          msg_id, msg, ts_added_in_buff = to_send
          self._heavy_ops_manager.run_all_on_comm_thread(msg)
          # now add whatever you need in the message
          msg = self._prepare_message(msg=msg, msg_id=msg_id)
          # now convert and send the dict as json
          bytes_delivered = self.send_wrapper(msg)
          self.telemetry_maybe_add_message(msg=msg, ts_added_in_buff=ts_added_in_buff, successful_send=bytes_delivered > 0)
          if bytes_delivered <= 0:
            self.P("Failed to send message: {}...".format(
              msg.get(ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, "<Unknown path>")
              ), color='r'
            )
          elif bytes_delivered == 1:
            self.P("Failed to send message: {} most likely due to large size...".format(
              msg.get(ct.PAYLOAD_DATA.EE_PAYLOAD_PATH, "<Unknown path>")
              ), color='r'
            )
          #endif
            
        #endif
        end_it = time()
        loop_time = end_it - start_it
        # loop_resolution = max(2 * len(self._send_buff), 10)
        loop_resolution = self.loop_resolution
        sleep_time = max(1 / loop_resolution - loop_time, 0.00001)
        sleep(sleep_time)
        self.loop_timings.append(loop_time)
      except Exception as e:
        msg = "Exception in default comm thread loop. Forcing loop delay extension until the exception does not occur."
        self.P(msg, color='r')
        sleep(5)
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,
          msg=msg,
          autocomplete_info=True
        )
      #end try-except
    #endwhile

    self._release()
    self.P('`run_thread` finished')
    self._thread_stopped = True
    return

